/*  detect.c
 *
 *  Signal detection functions of xfhell application
 */

/*
 *  xfhell: An application to transmit and receive
 *  Feld Hell signals using a computer's sound card
 *
 *  Copyright (C) 2003  Neoklis Kyriazis
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */

#include "xfhell.h"

/* Runtime config data */
extern rc_data_t rc_data;

/*------------------------------------------------------------------------*/

/*  Get_Element()
 *
 *  Gets the signal level over a part of a character's dot.
 */

  int
Get_Element( void )
{
  int elem_lev;

  if( isFlagSet(MODE_FELDHELL) )
	elem_lev = Feld_Level();
  else
	elem_lev = FMHell_Level();

  /* Display (scaled) signal level graph */
  if( isFlagSet(ENABLE_SCOPE) )
	Display_Signal( elem_lev );

  return( elem_lev );

} /* End of Get_Element() */

/*------------------------------------------------------------------------*/

/*  Feld_Level()
 *
 *  Detects level of a FeldHell signal using Goertzel's algorithm.
 */

  int
Feld_Level( void )
{
  static int
	first_call = TRUE,
	buff_idx; /* Sample buffer index */

  int
	idx,
	elem_lev,   /* Level of the FeldHell signal 'element'  */
	elem_timer; /* Counter for timing duration of element  */

  /* Circular signal samples buffer for Goertzel detector */
  static char *samples_buff = NULL;

  /* Variables for the Goertzel algorithm */
  static double cosw, sinw, coeff;
  double q0, q1, q2;


  /* Inititalize, first call or new baud rate */
  if( first_call || isFlagSet(NEW_BAUD_RATE) )
  {
	double w;

	first_call = FALSE;
	ClearFlag( NEW_BAUD_RATE );

	w = 2.0 * M_PI * (double)rc_data.tone_freq / (double)rc_data.dsp_speed;
	cosw  = cos(w);
	sinw  = sin(w);
	coeff = 2.0 * cosw;

	/* Allocate samples buffer */
	samples_buff = realloc(samples_buff,
		sizeof(char)*(size_t)rc_data.samples_perdot);
	if( samples_buff == NULL )
	{
	  perror( "samples_buffer" );
	  Error_Dialog(
		  "realloc() for samples_buffer failed\n"
		  "Quit xfhell and correct", TRUE );

	  ClearFlag( RECEIVE_MODE );
	  ClearFlag(
		  TRANSMIT_MODE  |
		  TRANSMIT_MACRO |
		  TRANSMIT_TAG   |
		  TRANSMIT_KEYBD );

	  return(-1);
	}
	bzero( samples_buff, (size_t)rc_data.samples_perdot);
	buff_idx = 0;

  } /* if( first_call || isFlagSet(NEW_BAUD_RATE) ) */

  /* Measure (screen) pixel level */
  for( elem_timer = 0; elem_timer < rc_data.samples_perpixel; elem_timer++ )
  {
	/* Get next signal sample from buffer, abort on error */
	samples_buff[buff_idx] = Signal_Sample();
	if( isFlagSet(DSP_IO_ERROR) )
	  return(-1);

	/* Icrement/reset circular buffers' index */
	if( ++buff_idx >= rc_data.samples_perdot )
	  buff_idx = 0;

  } /* for( elem_timer = 0; elem_timer < rc_data.samples_perpixel ... */

  /*** Calculate signal element level ***/
  /* Calculate element level using Goertzel algorithm */
  q1 = q2 = 0.0;
  for( idx = 0; idx < rc_data.samples_perdot; idx++ )
  {
	q0 = coeff * q1 - q2 + (double)samples_buff[buff_idx];
	q2 = q1;
	q1 = q0;

	/* Icrement/reset circular buffers' index */
	if( ++buff_idx >= rc_data.samples_perdot )
	  buff_idx = 0;
  }

  /* Level of input signal scaled */
  q1 /= (double)rc_data.samples_perdot;
  q2 /= (double)rc_data.samples_perdot;
  elem_lev = (int)( q1*q1 + q2*q2 - q1*q2*coeff )/16;
  if( elem_lev > 80 )
	elem_lev = 80;

  return( elem_lev );

} /* Feld_Level() */

/*------------------------------------------------------------------------*/

/* FMHell_Level()
 *
 * Detects frequency shifts of FMHell carrier
 */

  int
FMHell_Level( void )
{
  static int
	first_call = TRUE,
	det_period,
	buff_idx;   /* Sample buffer index */

  int
	idx,
	ret_val,
	blk_lev,    /* Level of the black FeldHell signal 'element' */
	wht_lev,    /* Level of the white FeldHell signal 'element' */
	elem_timer; /* Counter for timing duration of element */

  /* Circular signal samples buffer for Phase detector */
  static char *samples_buff = NULL;

  /* Variables for the Goertzel algorithm */
  static double blk_cosw, blk_sinw, blk_coeff, blk_scale;
  static double wht_cosw, wht_sinw, wht_coeff, wht_scale;
  double blk_q0, blk_q1, blk_q2;
  double wht_q0, wht_q1, wht_q2;


  /* Re-initialize on change of parameters */
  if( first_call || isFlagSet(NEW_BAUD_RATE) )
  {
	double w;

	det_period = rc_data.samples_perdot / rc_data.dot_resol;
	/* Actually w should be (tone +- 1/2baudrate)/dspspeed
	 * (half the effective baud rate in receive) but the
	 * full offset gives much better freq discrimination */
	w = 2.0 * M_PI *
	  ((double)rc_data.tone_freq + rc_data.baud_rate *
	   (double)rc_data.dot_resol / 2.0) / (double)rc_data.dsp_speed;
	wht_cosw  = cos(w);
	wht_sinw  = sin(w);
	wht_coeff = 2.0 * wht_cosw;
	wht_scale = w * det_period;

	w = 2.0 * M_PI *
	  ((double)rc_data.tone_freq - rc_data.baud_rate *
	   (double)rc_data.dot_resol / 2.0 ) / (double)rc_data.dsp_speed;
	blk_cosw  = cos(w);
	blk_sinw  = sin(w);
	blk_coeff = 2.0 * blk_cosw;
	blk_scale = w * det_period;

	/* Allocate samples buffer */
	samples_buff = realloc(samples_buff, sizeof(char)*(size_t)det_period);
	if( samples_buff == NULL )
	{
	  perror( "samples_buffer" );
	  Error_Dialog(
		  "realloc() for samples_buffer failed\n"
		  "Quit xfhell and correct", TRUE );

	  ClearFlag( RECEIVE_MODE );
	  ClearFlag(
		  TRANSMIT_MODE  |
		  TRANSMIT_MACRO |
		  TRANSMIT_TAG   |
		  TRANSMIT_KEYBD );

	  return(-1);
	}
	bzero( samples_buff, (size_t)det_period );
	buff_idx = 0;

	first_call = FALSE;
	ClearFlag( NEW_BAUD_RATE );

  } /* if( first_call... ) */

  /* Summate samples over each screen pixel */
  for(elem_timer = 0; elem_timer < rc_data.samples_perpixel; elem_timer++)
  {
	/* Get signal sample from buffer, abort on error */
	samples_buff[buff_idx] = Signal_Sample();
	if( isFlagSet(DSP_IO_ERROR) )
	{
	  puts(" DSP Error");
	  return(0);
	}

	/* Increment/reset circular buffers' index */
	if( ++buff_idx >= det_period )
	  buff_idx = 0;

  } /* for( elem_timer = 0; elem_timer < rc_data.samples_perpixel ... */

  /*** Calculate signal element level ***/
  /* Calculate element level using Goertzel algorithm */
  blk_q1 = blk_q2 = 0.0;
  wht_q1 = wht_q2 = 0.0;
  for( idx = 0; idx < det_period; idx++ )
  {
	blk_q0 = blk_coeff * blk_q1 - blk_q2 + (double)samples_buff[buff_idx];
	blk_q2 = blk_q1;
	blk_q1 = blk_q0;

	wht_q0 = wht_coeff * wht_q1 - wht_q2 + (double)samples_buff[buff_idx];
	wht_q2 = wht_q1;
	wht_q1 = wht_q0;

	/* Increment/reset circular buffers' index */
	if( ++buff_idx >= det_period )
	  buff_idx = 0;

  } /* for( idx = 0; idx < det_period; idx++ ) */

  /* Magnitude of black tone scaled by dot size and tone freq */
  blk_q1 /= blk_scale;
  blk_q2 /= blk_scale;
  blk_lev = (int)( (blk_q1*blk_q1 + blk_q2*blk_q2 - blk_q1*blk_q2*blk_coeff) );

  /* Magnitude of white tone scaled by dot size and tone freq */
  wht_q1 /= wht_scale;
  wht_q2 /= wht_scale;
  wht_lev = (int)( (wht_q1*wht_q1 + wht_q2*wht_q2 - wht_q1*wht_q2*wht_coeff) );
  if( blk_lev > wht_lev )
	ret_val = 80;
  else
	ret_val = 0;

  return( ret_val );

} /* End of FMHell_Level() */

/*------------------------------------------------------------------------*/

