/*  display.c
 *
 *  Display/Rendering functions for xfhell
 */

/*
 *  xfhell: An application to transmit and receive
 *  Feld Hell signals using a computer's sound card
 *
 *  Copyright (C) 2003  Neoklis Kyriazis
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>
#include "xfhell.h"
#include "support.h"

/* xfhell main window */
extern GtkWidget *main_window;

/* Global widgets */
extern GtkWidget
  *gbl_scope,
  *gbl_drawingarea,
  *gbl_tx_scrolledwindow;

/* Receive hellschreiber window pixbuf */
extern GdkPixbuf *gbl_pixbuf;
extern guchar *gbl_pixels;
extern guint
  gbl_rowstride,
  gbl_n_channels,
  gbl_width,
  gbl_height;

/* Waterfall window pixbuf */
extern GdkPixbuf *gbl_wfall_pixbuf;
extern guchar *gbl_wfall_pixels;
extern gint
  gbl_wfall_rowstride,
  gbl_wfall_n_channels,
  gbl_wfall_width,
  gbl_wfall_height;

/* Text buffers for text viewers */
extern GtkTextBuffer *gbl_tx_text_buffer;

/* Index to macros */
int gbl_macro_idx;

/* Runtime config data */
extern rc_data_t rc_data;

/* Average bin values */
int gbl_bin_ave[FFT_SIZE/2];

/* fft in/out (real/imaginary) buffers */
extern int
  *fft_in_r,
  *fft_out_r,
  *fft_out_i;

/*------------------------------------------------------------------------*/

/*  Draw_Column()
 *
 *  Draws a character's column in the Rx drawing area
 */

  gboolean
Draw_Column( unsigned char *column )
{
  /* Index to current pixel in pixbuf */
  static guchar	*chr_pix;

  /* Char row height, drawing area column and row */
  static int
	scrn_fill,	/* How much of Rx screen is filled */
	row_hgt,	/* Total height of rows on screen */
	chr_col=0,	/* Current column when painting characters */
	chr_row;	/* Current row when painting characters */

  int
	idx, i, j,
	pix_idx; /* Index to pixel in column */

  static unsigned char *last_column = NULL;


  /* Clear Receive window */
  if( isFlagSet(CLEAR_RX_WINDOW) )
  {
	ClearFlag( CLEAR_RX_WINDOW );

	row_hgt = rc_data.num_rows * rc_data.bitmap_hgt;
	if( chr_col > 2 ) Save_Pixbuf();
	Clear_Pixbuf();
	chr_col = 2;
	scrn_fill = 0;

	/* Allocate/clear memory for column buffer */
	last_column = realloc( last_column,
		sizeof(unsigned char) * (size_t)rc_data.bitmap_hgt );
	if( last_column == NULL )
	{
	  perror( "last_column" );
	  Error_Dialog(
		  "realloc() for last_column failed\n"
		  "Quit xfhell and correct", TRUE );
	  ClearFlag( RECEIVE_MODE );
	  ClearFlag(
		  TRANSMIT_MODE  |
		  TRANSMIT_MACRO |
		  TRANSMIT_TAG   |
		  TRANSMIT_KEYBD );
	  return( FALSE );
	}
	memset( last_column, 0,
		sizeof(unsigned char) * (size_t)rc_data.bitmap_hgt );

	/* Put cursor and char line at bottom of display */
	chr_row = gbl_drawingarea->allocation.height-2 - row_hgt;
	chr_pix =
	  gbl_pixels +
	  (gbl_drawingarea->allocation.height-2 - row_hgt) * gbl_rowstride +
	  2 * gbl_n_channels;
  } /* if( isFlagSet(CLEAR_RX_WINDOW) ) */

  /*** Draw a character column ***/
  /* Number of columns depends on dot size: */
  /* 1 for 1x2, and 2 and 2x4 pixels */
  for( idx = 0; idx < rc_data.dot_wdt; idx++ )
  {
	/* Draw elements, one per pixel, to pixbuf */
	for( pix_idx = 0; pix_idx < rc_data.bitmap_hgt; pix_idx++ )
	{
	  unsigned char grlev; /* dot gray level */

	  i = gbl_rowstride * (rc_data.bitmap_hgt - pix_idx - 1);
	  grlev = 0xff - column[pix_idx];
	  chr_pix[i  ] = grlev;
	  chr_pix[i+1] = grlev;
	  chr_pix[i+2] = grlev;
	}

	/* Draw 2nd raw if traditional 2-row mode is in use */
	if( isFlagClear(AUTO_DESKEW) )
	{
	  for( pix_idx = 0; pix_idx < rc_data.bitmap_hgt; pix_idx++ )
	  {
		unsigned char grlev; /* dot gray level */

		i = gbl_rowstride * (row_hgt - pix_idx - 1);
		grlev = 0xff - last_column[pix_idx];
		chr_pix[i  ] = grlev;
		chr_pix[i+1] = grlev;
		chr_pix[i+2] = grlev;
	  }
	}

	/* Draw new column */
	gtk_widget_queue_draw_area( gbl_drawingarea, chr_col, chr_row, 1, row_hgt );

	chr_col++;
	chr_pix += gbl_n_channels;

	/* Move back to beginning of row when reaching end of row */
	if( chr_col >= gbl_width-2 )
	{
	  chr_col = rc_data.bitmap_wdt + 2;
	  chr_pix = gbl_pixels + chr_col*gbl_n_channels + chr_row*gbl_rowstride;

	  /* Save pixbuf when screen filled */
	  scrn_fill += row_hgt;
	  if( scrn_fill >= gbl_drawingarea->allocation.height-4 )
	  {
		Save_Pixbuf();
		scrn_fill = 0;
	  }

	  /* Scroll up display one row */
	  gdk_pixbuf_copy_area(
		  gbl_pixbuf,
		  2, row_hgt+2,
		  gbl_drawingarea->allocation.width-4,
		  gbl_drawingarea->allocation.height-4 - row_hgt,
		  gbl_pixbuf, 2, 2 );

	  /* Redraw last block to beginning of new line */
	  gdk_pixbuf_copy_area(
		  gbl_pixbuf,
		  gbl_width-2 - rc_data.bitmap_wdt, chr_row,
		  rc_data.bitmap_wdt, row_hgt,
		  gbl_pixbuf, 2, chr_row );

	  /* Clear new row */
	  for( pix_idx = 0; pix_idx < row_hgt; pix_idx++ )
	  {
		i = (gbl_drawingarea->allocation.height-3 - pix_idx)*gbl_rowstride;
		i += chr_col * gbl_n_channels;
		for( j = chr_col; j < gbl_drawingarea->allocation.width-3; j++ )
		{
		  i += gbl_n_channels;
		  gbl_pixels[i  ] = 0xff;
		  gbl_pixels[i+1] = 0xff;
		  gbl_pixels[i+2] = 0xff;
		}
	  }

	  gtk_widget_queue_draw( gbl_drawingarea );

	} /* if( chr_col >= gbl_width-2 ) */

  } /* for( idx = 0; idx < rc_data.dot_wdt; idx++ ) */

  memcpy( last_column, column, (size_t)rc_data.bitmap_hgt );
  return( TRUE );

} /* Draw_Column() */

/*------------------------------------------------------------------------*/

/* Display_Waterfall()
 *
 * Displays audio spectrum as "waterfall"
 */

  void
Display_Waterfall(void)
{
  int
	idh, idv,  /* Index to hor. and vert. position in warterfall */
	pixel_val, /* Greyscale value of pixel derived from fft o/p  */
	fft_idx;   /* Index to fft output array */

  int
	bin_val,   /* Value of fft output "bin" */
	bin_max=0; /* Maximum value of fft bins */

  /* Sliding window average of bin max values */
  static int ave_max = 1000;

  /* Pointer to current pixel */
  static guchar *pix;


  /* Draw a vertical red line in waterfall at detector's freq. */
  pix = gbl_wfall_pixels + gbl_wfall_rowstride + gbl_wfall_n_channels;
  pix += gbl_wfall_n_channels * rc_data.red_line;
  pix[0] = 0xff;
  pix[1] = pix[2] = 0;

  /* Copy each line of waterfall to next one */
  for( idv = gbl_wfall_height-2; idv > 0; idv-- )
  {
	pix = gbl_wfall_pixels + gbl_wfall_rowstride * idv + gbl_wfall_n_channels;

	for( idh = 0; idh < WFALL_WIDTH; idh++ )
	{
	  pix[0] = pix[ -gbl_wfall_rowstride];
	  pix[1] = pix[1-gbl_wfall_rowstride];
	  pix[2] = pix[2-gbl_wfall_rowstride];
	  pix += gbl_wfall_n_channels;
	}
  }

  /* Got to top left +1 hor. +1 vert. of pixbuf */
  pix = gbl_wfall_pixels + gbl_wfall_rowstride + gbl_wfall_n_channels;

  /* First (DC) fft bin not used */
  gbl_bin_ave[0] = 0;

  /* Do the FFT on input array */
  Ifft();
  for( fft_idx = 0; fft_idx < FFT_SIZE/2; fft_idx++ )
  {
	/* Calculate vector magnitude of signal at each freq. ("bin") */
	fft_out_r[fft_idx] /= 8;
	fft_out_i[fft_idx] /= 8;
	bin_val =
	  fft_out_r[fft_idx]*fft_out_r[fft_idx] +
	  fft_out_i[fft_idx]*fft_out_i[fft_idx];

	/* Calculate average bin values */
	gbl_bin_ave[fft_idx] = ( (gbl_bin_ave[fft_idx] *
		  (BINMAX_AVE_WIN-1) + bin_val) / BINMAX_AVE_WIN );

	/* Record max bin value */
	if( bin_max < bin_val )
	  bin_max = bin_val;

	/* Scale pixel values to 255 depending on ave max value */
	pixel_val = (255 * bin_val) / ave_max;
	if( pixel_val > 255 )
	  pixel_val = 255;

	/* Duplicate pixels if fft size < waterfall width */
	for( idh = 0; idh < (2*WFALL_WIDTH)/(FFT_SIZE); idh++ )
	{
	  /* Reverse video is toggled by right click on waterfall */
	  if( isFlagSet(REVERSE_VIDEO) )
		pix[0] = pix[1] = pix[2] = pixel_val;
	  else
		pix[0] = pix[1] = pix[2] = 255 - pixel_val;

	  pix += gbl_wfall_n_channels;
	}

  } /* for( fft_idx = 0; fft_idx < fft_end; fft_idx++ ) */

  /* Calculate sliding window average of max bin value */
  ave_max = (ave_max*(BINMAX_AVE_WIN-1) + bin_max)/BINMAX_AVE_WIN;

  /* At last draw waterfall */
  gtk_widget_queue_draw( gbl_scope );

} /* Display_Waterfall() */

/*------------------------------------------------------------------------*/

/*  Clear_Pixbuf()
 *
 *  Initializes pixbuf of Receive window
 */

  void
Clear_Pixbuf( void )
{
  int i;
  guchar *p1, *p2;

  gdk_pixbuf_fill( gbl_pixbuf, 0xffffffff );

  /* Draw a box around pixbuf */
  p1 = gbl_pixels;
  p2 = gbl_pixels + (gbl_height-1)*gbl_rowstride;
  for( i = 0; i < gbl_width; i++ )
  {
	p1[0] = p1[1] = p1[2] = 0;
	p1 += gbl_n_channels;
	p2[0] = p2[1] = p2[2] = 0;
	p2 += gbl_n_channels;
  }

  p1 = gbl_pixels+gbl_rowstride;
  p2 = gbl_pixels+gbl_rowstride + (gbl_width-1)*gbl_n_channels;
  for( i = 0; i < gbl_height-1; i++ )
  {
	p1[0] = p1[1] = p1[2] = 0;
	p1 += gbl_rowstride;
	p2[0] = p2[1] = p2[2] = 0;
	p2 += gbl_rowstride;
  }

  gtk_widget_queue_draw( gbl_drawingarea );

} /* Clear_Pixbuf() */


/*------------------------------------------------------------------------*/

/*  Clear_Wfall_Pixbuf()
 *
 *  Initializes pixbuf of Receive window
 */

  void
Clear_Wfall_Pixbuf( void )
{
  int i;
  guchar *p1, *p2;

  gdk_pixbuf_fill( gbl_wfall_pixbuf, 0xffffffff );

  /* Draw a box around pixbuf */
  p1 = gbl_wfall_pixels;
  p2 = gbl_wfall_pixels + (gbl_wfall_height-1)*gbl_wfall_rowstride;
  for( i = 0; i < gbl_wfall_width; i++ )
  {
	p1[0] = p1[1] = p1[2] = 0;
	p1 += gbl_wfall_n_channels;
	p2[0] = p2[1] = p2[2] = 0;
	p2 += gbl_wfall_n_channels;
  }

  p1 = gbl_wfall_pixels+gbl_wfall_rowstride;
  p2 = gbl_wfall_pixels+gbl_wfall_rowstride +
	(gbl_wfall_width-1)*gbl_wfall_n_channels;
  for( i = 0; i < gbl_wfall_height-1; i++ )
  {
	p1[0] = p1[1] = p1[2] = 0;
	p1 += gbl_wfall_rowstride;
	p2[0] = p2[1] = p2[2] = 0;
	p2 += gbl_wfall_rowstride;
  }

  gtk_widget_queue_draw( gbl_scope );

} /* Clear_Wfall_Pixbuf() */

/*------------------------------------------------------------------------*/

/*  Print_Character()
 *
 *  Prints a character to a text view port
 */

void
Print_Character(
	int printchr,
	GtkTextBuffer *text_buffer,
	GtkWidget     *scrolledwindow,
	GtkTextIter   *iter )
{
  GtkAdjustment *adjustment;
  char char2text[2];

  /* Delete a char on backspace */
  if( printchr == BS )
  {
	GtkTextIter start;

	gtk_text_buffer_get_iter_at_offset(
		text_buffer, &start,
		gtk_text_buffer_get_char_count
		(text_buffer) -1 );
	gtk_text_buffer_delete( text_buffer, &start, iter );
	return;
  }

  /* Convert char to "text" */
  char2text[0] = (char) printchr;
  char2text[1] = '\0';

  /* Print character */
  gtk_text_buffer_insert(
	  text_buffer, iter,
	  char2text, -1 );

  /* Record to QSO record file */
  if( isFlagSet(RECORD_QSO) )
	fprintf( rc_data.qso_record_fp, "%c", printchr );

  /* Scroll Text View to bottom on LF */
  if( (printchr == LF) ||
	  (printchr == GDK_Return) )
  {
	adjustment = gtk_scrolled_window_get_vadjustment
	  ( GTK_SCROLLED_WINDOW(scrolledwindow) );
	gtk_adjustment_set_value( adjustment, adjustment->upper );
	if( isFlagSet(RECORD_QSO) )
	  fprintf( rc_data.qso_record_fp, "Tx-> " );
  }

} /* Print_Character() */

/*------------------------------------------------------------------------*/

/*  Display_Signal()
 *
 *  Updates the signal scope display */

  void
Display_Signal( int plot )
{
  /* Points to plot */
  static GdkPoint *points = NULL;
  static int
	first_call = TRUE,
	points_idx = 0;


  /* Initialize on first call */
  if( first_call )
  {
	first_call = FALSE;
	points = malloc( sizeof(GdkPoint) * (gbl_wfall_width-2) );
  }

  /* Save values to be plotted */
  points[points_idx].y = gbl_wfall_height-2 - plot;
  points[points_idx].x = points_idx + 1;

  /* Recycle buffer idx when full and plot */
  if( ++points_idx == gbl_wfall_width-2 )
  {
	/* Draw scope background */
	gdk_draw_rectangle( gbl_scope->window,
		gbl_scope->style->white_gc,
		TRUE, 0, 0,
		gbl_wfall_width-1, gbl_wfall_height-1 );

	/* Draw scope frame */
	gdk_draw_rectangle( gbl_scope->window,
		gbl_scope->style->black_gc,
		FALSE, 0, 0,
		gbl_wfall_width-1, gbl_wfall_height-1 );

	/* Plot signal graph */
	gdk_draw_lines( gbl_scope->window,
		gbl_scope->style->black_gc,
		points, gbl_wfall_width-2 );

	points_idx = 0;

  } /* if( ++points_idx == SCOPE_LEN-4 ) */

} /* Display_Signal( void ) */

/*------------------------------------------------------------------------*/
