/*  fft.c
 *
 *  FFT functions of xfhell application
 */

/*
 *  xfhell: An application to transmit and receive
 *  Feld Hell signals using a computer's sound card
 *
 *  Copyright (C) 2003  Neoklis Kyriazis
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */

#include "xfhell.h"

/* fft in/out buffers */
int
  *fft_in_r,
  *fft_out_r,
  *fft_out_i;

/* Scaled-up, integer, sin/cos tables */
static int
  isin[FFT_SIZE],
  icos[FFT_SIZE];

/*------------------------------------------------------------------------*/

/* Ifft_Init()
 *
 * Initializes fft()
 */
  void
Ifft_Init( void )
{
  int i;
  double w, tpi;

  /* Allocate fft buffers */
  i = sizeof(int) * FFT_SIZE;
  fft_in_r  = malloc( i );
  if( fft_in_r == NULL )
  {
	perror( "xfhell: fft_in_r malloc() failed");
	Error_Dialog(
		"fft_in_r malloc() failed\n"
		"Quit xfhell and correct", TRUE );
	return;
  }
  bzero( fft_in_r, i );

  i = sizeof(int) * FFT_SIZE/2;
  fft_out_r = malloc( i );
  if( fft_out_r == NULL )
  {
	perror( "xfhell: fft_out_r malloc() failed");
	Error_Dialog(
		"fft_out_r malloc() failed\n"
		"Quit xfhell and correct", TRUE );
	return;
  }

  fft_out_i = malloc( i );
  if( fft_out_i == NULL )
  {
	perror( "xfhell: fft_out_i malloc() failed");
	Error_Dialog(
		"fft_out_i malloc() failed\n"
		"Quit xfhell and correct", TRUE );
	return;
  }

  /* Make sin/cos tables */
  tpi = 2.0*M_PI;
  for( i = 0; i < FFT_SIZE; i++ )
  {
	w = tpi * (double)i / (double)FFT_SIZE;
	isin[i] = (int)(32.0 * sin(w) + 0.5);
	icos[i] = (int)(32.0 * cos(w) + 0.5);
  }

} /* Ifft_Init() */

/*------------------------------------------------------------------------*/

/* Ifft()
 *
 * Simple, integer-only, FFT function
 */
  void
Ifft( void )
{
  int i, j, w;
  /* In-phase and quadrature summation */
  int sum_i, sum_q;
  static int k = FFT_SIZE/2;

  /* Calculate output bins */
  for( i = 0; i < k; i++ )
  {
	sum_i = 0;
	sum_q = 0;

	/* Summate input values */
	for( j = 0; j < FFT_SIZE; j++ )
	{
	  w = (i*j) % FFT_SIZE;
	  sum_i += fft_in_r[j] * isin[w];
	  sum_q += fft_in_r[j] * icos[w];
	}

	/* Normalized summations to bins */
	fft_out_r[i] = sum_i/FFT_SIZE;
	fft_out_i[i] = sum_q/FFT_SIZE;
  }

} /* Ifft() */

/*------------------------------------------------------------------------*/

