/* operation.c
 *
 *  Operating modes functions for xfhell
 */

/*
 *  xfhell: An application to transmit and receive
 *  Feld Hell signals using a computer's sound card.
 *
 *  Copyright (C) 2003  Neoklis Kyriazis
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>
#include "xfhell.h"
#include "support.h"

/* xfhell main window */
extern GtkWidget *main_window;

/* Global widgets */
extern GtkWidget
  *gbl_tx_scrolledwindow,
  *gbl_drawingarea;

/* Receive window pixbuf */
extern GdkPixbuf *gbl_pixbuf;
extern guchar *gbl_pixels;
extern gint
  gbl_rowstride,
  gbl_n_channels,
  gbl_width,
  gbl_height;

/* Text buffer for Tx text viewer */
extern GtkTextBuffer *gbl_tx_text_buffer;

/* Keyboard entry buffer */
extern guint keybd_buff[32];

extern int
  newc_ptr,  /* Keystrokes counter */
  txch_ptr,  /* Pointer to character for transmission */
  space_cnt; /* Count of spaces typed in */

/* QSO data record */
extern qso_record_t qso_record;

extern char
  label[9][17], /* Labels for F1-F9 commands */
  *macro[9];    /* Macros attached to F1-F9  */

/* DSP buffer */
extern unsigned char *dsp_buff;

/* Index to macros */
int gbl_macro_idx;

/* Runtime config data */
extern rc_data_t rc_data;


/*------------------------------------------------------------------------*/

  gboolean
Transmit_Keybd( gpointer data )
{
  /* First call of function flag */
  static gboolean first_call = TRUE;


  /*** Cancel Transmit if Tx flag cleared ***/
  if( isFlagClear(TRANSMIT_KEYBD) &&
	  isFlagClear(KEYBD_BUSY) )
  {
	if( isFlagClear(TRANSMIT_MACRO) )
	  ClearFlag( TRANSMIT_MODE );
	first_call = TRUE;
	return( FALSE );

  } /* if( isFlagClear(TRANSMIT_KEYBD) ) */

  /*** Initialize if needed ***/
  if( first_call )
  {
	first_call = FALSE;
	gtk_widget_grab_focus( lookup_widget(main_window, "tx_textview") );
	if( isFlagClear(TRANSMIT_MODE) )
	  if( !Initialize_Transmit(data) )
	  {
		first_call = TRUE;
		return( FALSE );
	  }

	return( TRUE );
  } /* if( first_call ) */

  /* Transmit key buffer */
  if( space_cnt > 0 )
  {
	/* Xmit "smiley" etc glyphs with F1-F12 */
	if( (keybd_buff[txch_ptr] >= GDK_F1) &&
		(keybd_buff[txch_ptr] <= GDK_F12) )
	{
	  if( ! Transmit_Character(&rc_data.font_data,
			16 + keybd_buff[txch_ptr] - GDK_F1) )
		return( FALSE );
	}
	else
	{
	  /* Illegal chars are ignored */
	  if( (keybd_buff[txch_ptr] >= rc_data.font_data.first_glyph) &&
		  (keybd_buff[txch_ptr] <= rc_data.font_data.last_glyph) )
		if( ! Transmit_Character(&rc_data.font_data, keybd_buff[txch_ptr]) )
		  return( FALSE );
	}

	/* Count down spaces */
	if( keybd_buff[txch_ptr] == ' ' )
	  space_cnt--;

	/* Reset key buff pointer */
	if( ++txch_ptr == KEY_BUFF_SIZE )
	  txch_ptr = 0;

  } /* if( space_cnt > 0 ) */
  else
	/* Transmit idle character */
  {
	if( ! Transmit_Character(&rc_data.font_data, IDLE_CHAR) )
	  return( FALSE );

	ClearFlag( KEYBD_BUSY );

	if( isFlagSet(TRANSMIT_MACRO) )
	  ClearFlag( TRANSMIT_KEYBD );
	else
	  if( isFlagClear(TRANSMIT_KEYBD) )
		ClearFlag( TRANSMIT_MODE );
  }

  return( TRUE );

} /* End of Transmit_Keybd() */

/*------------------------------------------------------------------------*/

/*  Transmit_Macro()
 *
 *  Transmits a pre-recorded Macro
 */

  gboolean
Transmit_Macro( gpointer data )
{
  /* Text buffer marker */
  static GtkTextIter iter;

  static int
	macro_cnt = 0, /* Count macro characters tx'd */
	tag_idx   = 0; /* Index for transmitting tags */

  /* First call of function flag */
  static gboolean first_call = TRUE;

  /* Tags for entering fields in macros */
  char *tags[NUM_OF_TAGS] = TAG_TABLE;

  /* String within a tag <> */
  char tag_str[10];

  /* Tag replacement string */
  static char tag_rep[22];


  /* Delay macro till keybd not busy */
  if( isFlagSet(KEYBD_BUSY) )
	return( TRUE );

  /* Abort macro if user disabled, remain in Tx mode */
  /* Stop macro on '~' character, remain in Tx mode  */
  if( isFlagClear(TRANSMIT_MACRO) ||
	  (macro[gbl_macro_idx][macro_cnt] == '~') )
  {
	if( ! Transmit_Character(&rc_data.font_data, IDLE_CHAR) )
	  return( FALSE );

	if( isFlagSet(RECORD_QSO) )
	  fflush( rc_data.qso_record_fp );

	macro_cnt = tag_idx = 0;
	first_call = TRUE;
	ClearFlag( TRANSMIT_MACRO |	TRANSMIT_TAG );
	SetFlag( TRANSMIT_KEYBD );
	gtk_idle_add( Transmit_Keybd, NULL );

	return( FALSE );
  }

  /* Stop macro at string terminator  */
  /* or CW id char (%), go to Rx mode */
  if( (macro[gbl_macro_idx][macro_cnt] == '\0') ||
	  (macro[gbl_macro_idx][macro_cnt] == '*') )
  {
	if( isFlagSet(DSP_IO_ERROR) )
	  return( FALSE );

	if( isFlagSet(RECORD_QSO) )
	  fflush( rc_data.qso_record_fp );

	/* Transmit Morse ID */
	if( macro[gbl_macro_idx][macro_cnt] == '*' )
	{
	  Morse_Transmit( rc_data.cw_msg );
	  if( isFlagSet(DSP_IO_ERROR) )
		return( FALSE );
	}

	macro_cnt = tag_idx = 0;
	first_call = TRUE;
	ClearFlag(
		TRANSMIT_MACRO |
		TRANSMIT_KEYBD |
		TRANSMIT_MODE  |
		TRANSMIT_TAG );
	SetFlag( RECEIVE_MODE );
	gtk_idle_add( Receive_Mode, NULL );
	return( FALSE );
  }

  /*** Initialize if needed ***/
  if( first_call )
  {
	first_call = FALSE;
	if( isFlagClear(TRANSMIT_MODE) )
	  if( !Initialize_Transmit(data) )
	  {
		first_call = TRUE;
		return( FALSE );
	  }

	return( TRUE );
  } /* if( first_call ) */

  /* Set up TRANSMIT text view */
  gtk_text_buffer_get_iter_at_offset( gbl_tx_text_buffer, &iter,
	  gtk_text_buffer_get_char_count
	  (gbl_tx_text_buffer) );

  /*** Look for tags and substitude accordingly ***/
  if( (macro[gbl_macro_idx][macro_cnt] == '<') &&
	  isFlagClear(TRANSMIT_TAG) )
  {
	int idx = 0;

	SetFlag( TRANSMIT_TAG );

	/*** Copy tag string to buffer and identify ***/
	macro_cnt++; /* Move into tag field */

	/* Copy line to tag, max 12 chars or EOS */
	while( (idx < 12) &&
		(macro[gbl_macro_idx][macro_cnt] != '>') )
	  tag_str[idx++] = macro[gbl_macro_idx][macro_cnt++];

	/* Terminate tag buffer and advance macro buffer */
	tag_str[idx] = '\0';
	macro_cnt++;

	/* Identify tag and substitude accordingly */
	idx = 0;
	while( idx < NUM_OF_TAGS )
	{
	  if( strcmp(tag_str, tags[idx]) == 0 )
		break;
	  idx++;
	}

	/* Abort if tag unidentified */
	if( idx == NUM_OF_TAGS )
	{
	  /* If tag is not a number */
	  if( (tag_str[0] < 0x30) && (tag_str[0] > 0x39) )
	  {
		Error_Dialog(
			"Error reading xfhellrc file\n"
			"A tag in a macro is invalid\n"
			"Quit xfhell and correct", TRUE );
		return( FALSE );
	  }
	  else
	  {
		/* Take number as glyph encoding and xmit */
		if( ! Transmit_Character(&rc_data.font_data, atoi(tag_str)) )
		  return( FALSE );
		ClearFlag( TRANSMIT_TAG );
		return( TRUE );
	  }
	}

	/* Replace tags */
	switch( idx )
	{
	  case 0: /* own-call */
		strncpy( tag_rep, rc_data.call, 17 );
		break;

	  case 1: /* own-name */
		strncpy( tag_rep, rc_data.name, 17 );
		break;

	  case 2: /* own-qth */
		strncpy( tag_rep, rc_data.qth, 21 );
		break;

	  case 3: /* own-loc */
		strncpy( tag_rep, rc_data.loc, 7 );
		break;

	  case 4: /* own-rst */
		strncpy( tag_rep, qso_record.my_rst, 4 );
		break;

	  case 5: /* rem-call */
		strncpy( tag_rep, qso_record.dx_call, 15 );
		break;

	  case 6: /* rem-name */
		strncpy( tag_rep, qso_record.dx_name, 13 );
		break;

	  case 7: /* rem-qth */
		strncpy( tag_rep, qso_record.dx_qth, 13 );
		break;

	  case 8: /* rem-loc */
		strncpy( tag_rep, qso_record.dx_loc, 7 );
		break;

	  case 9: /* rem-rst */
		strncpy( tag_rep, qso_record.dx_rst, 4 );
		break;

	  case 10: /* date-time */
		strncpy( tag_rep, qso_record.date, 12 );
		strcat ( tag_rep, " " );
		strncat( tag_rep, qso_record.time, 9 );
		break;

	  case 11: /* op-freq, strip trailing spaces */
		{
		  int i;

		  strncpy( tag_rep, qso_record.freq, 14 );
		  i = strlen( tag_rep ) - 1;
		  for( ; i >= 0; i-- )
			if( tag_rep[i] == ' ' )
			  tag_rep[i] = '\0';
		}
		break;

	  case 12: /* app-version */
		snprintf( tag_rep, 22, "%s %s", PACKAGE, VERSION );

	} /* switch( idx ) */

  } /* if( (macro[gbl_macro_idx][macro_cnt] == '<') && ) */

  /*** Transmit tag replacement ***/
  if( isFlagSet(TRANSMIT_TAG) )
  {
	if( tag_rep[tag_idx] != '\0' )
	{
	  /* Capitalize letters if enabled */
	  if( isFlagSet(CAPITALIZE) &&
		  (tag_rep[tag_idx] > 0x60)  &&
		  (tag_rep[tag_idx] < 0x7b) )
		tag_rep[tag_idx] -= 0x20;

	  if( ! Transmit_Character(&rc_data.font_data, tag_rep[tag_idx]) )
		return( FALSE );
	  Print_Character( tag_rep[tag_idx], gbl_tx_text_buffer,
		  gbl_tx_scrolledwindow, &iter );
	  tag_idx++;

	} /* if( (tag_rep[tag_idx] != '\0') ) */
	else
	{
	  ClearFlag( TRANSMIT_TAG );
	  tag_idx = 0;
	}

	return( TRUE );

  } /* if( isFlagSet(TRANSMIT_TAG) && */

  /* Capitalize letters if enabled */
  if( isFlagSet(CAPITALIZE) &&
	  (macro[gbl_macro_idx][macro_cnt] > 0x60) &&
	  (macro[gbl_macro_idx][macro_cnt] < 0x7b) )
	macro[gbl_macro_idx][macro_cnt] -= 0x20;

  /* Transmit a char from Macro */
  if( (macro[gbl_macro_idx][macro_cnt] == CR) ||
	  (macro[gbl_macro_idx][macro_cnt] == LF) )
  {
	if( ! Transmit_Character(&rc_data.font_data, ' ') )
	  return( FALSE );
  }
  else
	if( ! Transmit_Character( &rc_data.font_data, macro[gbl_macro_idx][macro_cnt]) )
	  return( FALSE );

  Print_Character( macro[gbl_macro_idx][macro_cnt],
	  gbl_tx_text_buffer, gbl_tx_scrolledwindow, &iter );
  macro_cnt++;

  return( TRUE );

} /* Transmit_Macro() */

/*------------------------------------------------------------------------*/

  gboolean
Receive_Mode( gpointer data )
{
  /* First call of function flag */
  static gboolean first_call = TRUE;

  /* Character column data ("element" values) */
  unsigned char column[rc_data.bitmap_hgt];


  /* Delay receive till keybd not busy */
  if( isFlagSet(KEYBD_BUSY) )
	return( TRUE );

  /* Cancel Receive if Rx flag cleared */
  if( isFlagClear(RECEIVE_MODE) )
  {
	first_call = TRUE;
	return( FALSE );
  }

  /*** Initialize ***/
  if( first_call )
  {
	first_call = FALSE;

	/* Setup sound card, abort on error */
	Setup_Sound_Card();
	if( isFlagClear(SOUND_SETUP) )
	{
	  ClearFlag( RECEIVE_MODE );
	  Set_TxRx_Labels();
	  first_call = TRUE;
	  return( FALSE );
	}

	/* Set tcvr mode, abort on error */
	if( isFlagSet(ENABLE_CAT) )
	{
	  Set_Tcvr_Mode();
	  if( isFlagSet(SERIAL_IO_ERROR) )
	  {
		ClearFlag( RECEIVE_MODE );
		Set_TxRx_Labels();
		first_call = TRUE;
		return( FALSE );
	  }
	}

	/* Abort if QSO record file fails to open */
	if( !Open_Record_File() )
	{
	  first_call = TRUE;
	  return( FALSE );
	}

	/* Enable pixbuf save */
	SetFlag( SAVE_PIXBUF );

	Set_TxRx_Labels();

	return( TRUE );

  } /* if( first_call ) */

  /*** Draw characters to Receive window ***/

  /* Make and draw a column, abort on dsp error */
  Assemble_Column( column );
  if( isFlagSet(DSP_IO_ERROR) )
	return( FALSE );
  if( ! Draw_Column(column) )
	return( FALSE );

  return( TRUE );

} /* End of Receive_Mode() */

/*------------------------------------------------------------------------*/

/*  Initialize_Transmit()
 *
 *  Initializes the Transmit mode
 */

  gboolean
Initialize_Transmit(gpointer data)
{
  int idx;

  /* Setup sound card, abort on error */
  Setup_Sound_Card();
  if( isFlagClear(SOUND_SETUP) )
  {
	ClearFlag(
		TRANSMIT_KEYBD |
		TRANSMIT_MACRO |
		TRANSMIT_MODE );
	Set_TxRx_Labels();
	return( FALSE );
  }

  /* Clear buffer */
  for( idx = 0; idx < rc_data.txbuff_len; idx++ )
	dsp_buff[idx] = 0x80;

  /* Switch to USB and PTT ON, enable pixbuf save */
  if( isFlagSet(ENABLE_CAT) )
  {
	SetFlag( TRANSMIT_MODE );
	Set_Tcvr_Mode();
	if( isFlagSet(SERIAL_IO_ERROR) )
	{
	  ClearFlag( TRANSMIT_MODE );
	  Set_TxRx_Labels();
	  return( FALSE );
	}
  } /* if( isFlagSet(ENABLE_CAT) ) */

  Set_TxRx_Labels();

  return( Open_Record_File() );

} /* Initialize_Transmit() */

/*------------------------------------------------------------------------*/

