/*  sound.c
 *
 *  Soundcard handling functions of xfhell application
 */

/*
 *  xfhell: An application to transmit and receive
 *  Feld Hell signals using a computer's sound card
 *
 *  Copyright (C) 2003  Neoklis Kyriazis
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */

#include "xfhell.h"

extern int
  rx_buff_idx,	/* Index to signal samples buffer for Rx */
  tx_buff_idx,	/* Index to signal samples buffer for tx */
  resample;	/* Resample request in DSP samples */

/* Signal/DSP samples buffer. Double size for Stereo */
extern unsigned char *dsp_buff;

/* Config data */
extern rc_data_t rc_data;

/* fft in/out buffers */
extern int
  *fft_in_r,
  *fft_out_r,
  *fft_out_i;

/* Mixer device to be set-up (input or reclev) */
unsigned int rec_device;

/*------------------------------------------------------------------------*/

/*  Setup_Sound_Card()
 *
 *  Sets up mixer and DSP devices
 */

  void
Setup_Sound_Card( void )
{
  int
	new_recsrc, /* New recording source (RECORD_SRC) */
	level,      /* Rec/Volume level request to mixer */
	itmp;

  /* Error messages */
  char message[90];

  /* Sound device names from souncard.h */
  char *dev_name[SOUND_MIXER_NRDEVICES] = SOUND_DEVICE_NAMES;

  /* File lock information */
  struct flock lockinfo;


  /* Return if sound card is set-up */
  if( isFlagSet(SOUND_SETUP) )
	return;

  rc_data.dsp_in_fd = rc_data.dsp_out_fd = rc_data.mix_fd = 0;

  /*** Open mixer device, abort on error ***/
  if( (rc_data.mix_fd = open(rc_data.mixer, O_RDWR, 0)) == -1 )
  {
	perror( rc_data.mixer );
	snprintf( message, 90,
		"Unable to open %s\n"
		"Quit xfhell and correct", rc_data.mixer);
	Error_Dialog( message, TRUE );
	return;
  }

  /* Attempt to lock entire mixer device file */
  lockinfo.l_type   = F_WRLCK;
  lockinfo.l_whence = SEEK_SET;
  lockinfo.l_start  = 0;
  lockinfo.l_len    = 0;

  /* If mixer device is already locked, abort */
  if( fcntl( rc_data.mix_fd, F_SETLK, &lockinfo ) < 0 )
  {
	perror( rc_data.mixer );
	fcntl( rc_data.mix_fd, F_GETLK, &lockinfo );
	fprintf( stderr, "xfhell: Lock %s: Device is locked by pid %d\n",
		rc_data.mixer, lockinfo.l_pid );
	snprintf( message, 90,
		"Unable to lock %s\n"
		"Device locked by pid %d\n"
		"Quit xfhell and correct",
		rc_data.mixer, lockinfo.l_pid );
	Error_Dialog( message, TRUE );
	return;
  }

  /*** Setup mixer parameters ***/
  /* Find number of recording source matching source name */
  for( new_recsrc = 0; new_recsrc < SOUND_MIXER_NRDEVICES; new_recsrc++ )
	if( strcmp( dev_name[new_recsrc], rc_data.rec_src ) == 0 )
	  break;

  /* Make recording source mask */
  itmp = 1 << new_recsrc;

  /* Attempt to select recording source, abort on error */
  if( ioctl(rc_data.mix_fd, SOUND_MIXER_WRITE_RECSRC, &itmp) == -1 )
  {
	perror("xfhell: SOUND_MIXER_WRITE_RECSRC");
	snprintf( message, 90,
		"Failed to select Recording Source %s\n"
		"Quit xfhell and correct", rc_data.rec_src );
	Error_Dialog( message, TRUE );
	return;
  }
  if( itmp != 1 << new_recsrc )
  {
	snprintf( message, 90,
		"Failed to select Recording Source %s\n"
		"Quit xfhell and correct", rc_data.rec_src );
	Error_Dialog( message, TRUE );
	return;
  }

  /* Zero line level (to silence my speakers) */
  level = itmp = 0;
  if( ioctl(rc_data.mix_fd, MIXER_WRITE(new_recsrc), &level) == -1 )
  {
	perror("xfhell: Setting Recording source level");
	snprintf( message, 90,
		"Failed to set %s level\n"
		"Quit xfhell and correct", rc_data.rec_src );
	Error_Dialog( message, TRUE );
	return;
  }
  if( level != itmp )
  {
	snprintf( message, 90,
		"Failed to set %s level\n"
		"Quit xfhell and correct", rc_data.rec_src );
	Error_Dialog( message, TRUE );
	return;
  }

  /* Find number of recording level device */
  for( rec_device = 0; rec_device < SOUND_MIXER_NRDEVICES; rec_device++ )
	if( strcmp( dev_name[rec_device], rc_data.lev_dev ) == 0 )
	  break;
  if( rec_device ==  SOUND_MIXER_NRDEVICES )
  {
	snprintf( message, 90,
		"Failed to select Recording Level device %s\n"
		"Quit xfhell and correct", rc_data.lev_dev );
	Error_Dialog( message, TRUE );
	return;
  }

  /* Write level to Volume, abort on error */
  Set_Recording_Level( rc_data.rec_lev );
  Set_Volume_Level( rc_data.vol_lev );

  /*** Open and setup dsp device(s) ***/
  /* If only one device specified, open input dsp r/w */
  if( strcmp(rc_data.dsp_in, rc_data.dsp_out) == 0 )
  {
	if( !Setup_Dsp_Dev(rc_data.dsp_in, O_RDWR, &rc_data.dsp_in_fd) )
	  return;
	rc_data.dsp_out_fd = rc_data.dsp_in_fd;
  }
  else /* Setup separate dsp in/out */
  {
	if( !Setup_Dsp_Dev(rc_data.dsp_in, O_RDONLY, &rc_data.dsp_in_fd) )
	  return;

	if( !Setup_Dsp_Dev(rc_data.dsp_out, O_WRONLY, &rc_data.dsp_out_fd) )
	  return;
  }

  SetFlag( SOUND_SETUP );

} /* End of Setup_Sound_Card() */

/*------------------------------------------------------------------------*/

/*  Setup_Dsp_Dev()
 *
 *  Sets up a given dsp device
 */

  gboolean
Setup_Dsp_Dev( char *dsp_dev, int mode, int *dsp_fd )
{
  int
	dsp_speed,  /* DSP sampling speed as in xfhell.h */
	seg_size,   /* DSP buffer (segment) size in kb   */
	sample_fmt, /* DSP sample format, 8-bit unsigned */
	itmp;

  /* Error messages */
  char message[90];

  /* File lock information */
  struct flock lockinfo;


  /* Open input DSP device, abort on error */
  if( (*dsp_fd = open(dsp_dev, mode, 0)) == -1 )
  {
	perror( dsp_dev );
	snprintf( message, 90,
		"Unable to open %s\n"
		"Quit xfhell and correct", dsp_dev);
	Error_Dialog( message, TRUE );
	return FALSE;
  }

  /* Attempt to lock entire DSP device file */
  if( mode == O_RDONLY )
  {
	lockinfo.l_type   = F_RDLCK;
	lockinfo.l_whence = SEEK_SET;
	lockinfo.l_start  = 0;
	lockinfo.l_len    = 0;
  }
  else
	if( mode == O_WRONLY )
	{
	  lockinfo.l_type   = F_WRLCK;
	  lockinfo.l_whence = SEEK_SET;
	  lockinfo.l_start  = 0;
	  lockinfo.l_len    = 0;
	}
	else
	{
	  lockinfo.l_type   = F_RDLCK | F_WRLCK;
	  lockinfo.l_whence = SEEK_SET;
	  lockinfo.l_start  = 0;
	  lockinfo.l_len    = 0;
	}


  /* If DSP device is already locked, abort */
  if( fcntl( *dsp_fd, F_SETLK, &lockinfo ) < 0 )
  {
	perror( dsp_dev );
	fcntl( *dsp_fd, F_GETLK, &lockinfo );
	fprintf( stderr, "xfhell: Lock %s: Device is locked by pid %d\n",
		dsp_dev, lockinfo.l_pid );
	snprintf( message, 90,
		"Unable to lock %s\n"
		"Device locked by pid %d\n"
		"Quit xfhell and correct",
		dsp_dev, lockinfo.l_pid );
	Error_Dialog( message, TRUE );
	return FALSE;
  }

  /* Set segment size according to stereo/mono mode */
  itmp = 1; seg_size = 0;
  while( (itmp << seg_size) != RX_BUFFER_SIZE ) seg_size++;
  itmp += rc_data.num_chan << 18;
  ioctl( *dsp_fd, SNDCTL_DSP_SETFRAGMENT, &seg_size );

  /* Set sample format to unsigned 8-bit, abort on error */
  sample_fmt = AFMT_U8;
  if( ioctl(*dsp_fd, SNDCTL_DSP_SETFMT, &sample_fmt) == -1 )
  {
	perror( "xfhell: SNDCTL_DSP_SETFMT");
	Error_Dialog(
		"Failed to set dsp sample format AFMT_U8\n"
		"Quit xfhell and correct", TRUE );
	return FALSE;
  }
  if( sample_fmt != AFMT_U8 )
  {
	Error_Dialog(
		"Failed to set dsp sample format AFMT_U8\n"
		"Quit xfhell and correct", TRUE );
	return FALSE;
  }

  /* Set stereo/mono mode, continue on error */
  itmp = rc_data.num_chan;
  if( ioctl(*dsp_fd, SNDCTL_DSP_CHANNELS, &rc_data.num_chan) == -1 )
	perror("SNDCTL_DSP_CHANNELS");

  /* Setup DSP sampling speed, abort on error */
  dsp_speed = rc_data.dsp_speed;
  if( ioctl(*dsp_fd, SNDCTL_DSP_SPEED, &dsp_speed) == -1 )
  {
	perror( "xfhell: SNDCTL_DSP_SPEED");
	snprintf( message, 90,
		"Failed to set DSP speed to %d\n"
		"Quit xfhell and correct", rc_data.dsp_speed );
	Error_Dialog( message, TRUE );
	return FALSE;
  }
  if( dsp_speed != rc_data.dsp_speed )
  {
	snprintf( message, 90,
		"Failed to set DSP speed to %d\n"
		"Quit xfhell and correct", rc_data.dsp_speed );
	Error_Dialog( message, TRUE );
	return FALSE;
  }

  return TRUE;

} /* Setup_Dsp_Dev */

/*------------------------------------------------------------------------*/

/* Set_Volume_Level()
 *
 * Sets Volume Control level
 */
  void
Set_Volume_Level( int level )
{
  int lev, itmp, err = 0;

  /* Write level to Volume, abort on error */
  rc_data.vol_lev = level;
  lev = itmp = level | (level << 8);
  if( ioctl(rc_data.mix_fd, SOUND_MIXER_WRITE_VOLUME, &lev) == -1 )
  {
	perror("xfhell: Setting Volume level");
	err = 1;
  }
  if( lev != itmp )
	err = 1;

  /* Write level to PCM, abort on error */
  level = itmp = rc_data.vol_lev | (rc_data.vol_lev << 8);
  if( ioctl(rc_data.mix_fd, SOUND_MIXER_WRITE_PCM, &level) == -1 )
  {
	perror("xfhell: Setting PCM level");
	err = 1;
  }
  if( level != itmp )
	err = 1;

  if( err )
	Error_Dialog(
		"Failed to set volume level\n"
		"Quit xfhell and correct", TRUE );

} /* Set_Volume_Level() */

/*------------------------------------------------------------------------*/

/* Set_Capture_Level()
 *
 * Sets Recording Control level
 */
  void
Set_Recording_Level( int level )
{
  int lev, itmp, err = 0;

  /* Write recording level to mixer, abort on error */
  rc_data.rec_lev = level;
  lev = itmp = level | level << 8;
  if( ioctl(rc_data.mix_fd, MIXER_WRITE(rec_device), &lev) == -1 )
  {
	perror("xfhell: Setting Recording level");
	err = 1;
  }
  if( lev != itmp )
	err = 1;

  if( err )
	Error_Dialog(
		"Failed to set Recording level\n"
		"Quit xfhell and correct", TRUE );

} /* Set_Recording_Level() */

/*------------------------------------------------------------------------*/

/*  Signal_Sample()
 *
 *  Gets the next DSP sample of the signal input.
 */

  int
Signal_Sample( void )
{
  int
	sample_val = 0, /* Signal sample value */
	sample_cnt;     /* Number of samples read from DSP */

  /* Counts num of samples after which a sample is */
  /* dropped or repeated to refine sampling rate   */
  static int
	xtra_sample_cnt = 0,
	fft_idx = 0, /* fft input buffer idx */
	cnt = 0,	 /* Count of calls to this function */
	flag = TRUE;

  /* Initialize samples bufer */
  if( flag )
  {
	flag = FALSE;

	/* Start buffer index according to stereo/mono mode */
	if( rc_data.num_chan == 2 ) /* Stereo */
	  rx_buff_idx = rc_data.chan_used;
	else
	  rx_buff_idx = 0;

	/* Read audio samples from DSP, abort on error */
	if( (sample_cnt = read(rc_data.dsp_in_fd, dsp_buff, rc_data.rxbuff_len)) == -1 )
	{
	  perror( "xfhell: dsp read()" );
	  Error_Dialog(
		  "Error reading from dsp device\n"
		  "Quit xfhell and correct", TRUE );
	  SetFlag( DSP_IO_ERROR );
	}
	if( sample_cnt != rc_data.rxbuff_len )
	{
	  Error_Dialog(
		  "Error reading from dsp device\n"
		  "Quit xfhell and correct", TRUE );
	  SetFlag( DSP_IO_ERROR );
	}
  }

  if( isFlagSet(RECEIVE_MODE) && isFlagClear(KEYBD_BUSY) )
  {
	/* Refill DSP samples buffer when needed */
	if( rx_buff_idx >= rc_data.rxbuff_len )
	{
	  /* Start buffer index according to stereo/mono mode */
	  if( rc_data.num_chan == 2 ) /* Stereo */
		rx_buff_idx = rc_data.chan_used;
	  else
		rx_buff_idx = 0;

	  /* Read audio samples from DSP, abort on error */
	  if( (sample_cnt = read(rc_data.dsp_in_fd, dsp_buff, rc_data.rxbuff_len)) == -1 )
	  {
		perror( "xfhell: dsp read()" );
		Error_Dialog(
			"Error reading from dsp device\n"
			"Quit xfhell and correct", TRUE );
		SetFlag( DSP_IO_ERROR );
	  }
	  if( sample_cnt != rc_data.rxbuff_len )
	  {
		Error_Dialog(
			"Error reading from dsp device\n"
			"Quit xfhell and correct", TRUE );
		SetFlag( DSP_IO_ERROR );
	  }

	} /* End of if( rx_buff_idx == buffer_size ) */

	/* Remove neutral value (0x80) from samples */
	sample_val = dsp_buff[rx_buff_idx] - 0x80;

	/* Do a virtual control of dsp speed */
	if( resample )
	{
	  /* Return extra sample if resample req positive */
	  if( (resample > 0 ) &&
		  (xtra_sample_cnt++ >= RESAMPLE_CONST/rc_data.dsp_speed/resample) )
	  {
		xtra_sample_cnt = 0;
		return( sample_val );
	  }

	  /* Drop a sample if resample req negative */
	  if( (resample < 0 ) &&
		  (xtra_sample_cnt++ >= -RESAMPLE_CONST/rc_data.dsp_speed/resample) )
	  {
		xtra_sample_cnt = 0;
		rx_buff_idx += rc_data.num_chan;
	  }

	} /* if( resample != 0 ) */

	/* Increment according to mono/stereo mode */
	rx_buff_idx += rc_data.num_chan;

  } /* if( isFlagSet(RECEIVE_MODE) ) */
  else
	if( isFlagSet(TX2RX_LOOPBACK) )
	{
	  sample_val = dsp_buff[tx_buff_idx] - 0x80;
	  tx_buff_idx += rc_data.num_chan;

	} /* if( isFlagSet(RECEIVE_MODE) ) */

  /* Display waterfall when input buffer full.
   * fft_stride samples are added for each input element */
  if( isFlagClear(ENABLE_SCOPE) )
  {
	if( cnt == 0 )
	  fft_in_r[fft_idx] = sample_val;
	else
	  fft_in_r[fft_idx] += sample_val;

	if( ++cnt == rc_data.fft_stride )
	{
	  fft_idx++;
	  cnt = 0;
	}

	if( fft_idx >= FFT_SIZE )
	{
	  fft_idx = 0;
	  Display_Waterfall();
	}

  } /* if( isFlagClear(ENABLE_SCOPE) ) */

  return( sample_val );

} /* End of Signal_Sample() */

/*------------------------------------------------------------------------*/

/*  DSP_Write()
 *
 *  Write a samples buffer to DSP
 */

  void
DSP_Write( unsigned char *buffer, int buff_size )
{
  int sample_cnt;

  /*** Write samples buffer to DSP or file ***/
  if( (sample_cnt = write(rc_data.dsp_out_fd, buffer, buff_size)) == -1 )
  {
	perror( "xfhell: dsp read()" );
	Error_Dialog(
		"Error writing to dsp device\n"
		"Quit xfhell and correct", TRUE );
	SetFlag( DSP_IO_ERROR );
  }
  if( sample_cnt != buff_size )
  {
	Error_Dialog(
		"Error writing to dsp device\n"
		"Quit xfhell and correct", TRUE );
	SetFlag( DSP_IO_ERROR );
  }

} /* DSP_Write() */

/*------------------------------------------------------------------------*/
