/* utils.c
 *
 *  Utility functions for xfhell
 */

/*
 *  xfhell: An application to transmit and receive
 *  Feld Hell signals using a computer's sound card
 *
 *  Copyright (C) 2003  Neoklis Kyriazis
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>
#include "utils.h"
#include "support.h"

/* Signal samples buffer */
extern unsigned char *dsp_buff;

/* xfhell main window */
extern GtkWidget
  *main_window,
  *gbl_popup_menu;

/* Receive pix buff */
extern GdkPixbuf *gbl_pixbuf;

/* QSO data record */
extern qso_record_t qso_record;

/* Runtime config data */
extern rc_data_t rc_data;

extern char
  label[9][17], /* Labels for F1-F9 commands */
  *macro[9];    /* Macros attached to F1-F9  */

/* Index to macros */
extern int gbl_macro_idx;

/*------------------------------------------------------------------------*/

/*  Save_QSO_Record()
 *
 *  Saves QSO record to .adif and .txt files
 */

  gboolean
Save_QSO_Record( void )
{
  /* For opening files */
  char file_name[64];

  /* Buffer for reading and printing to files */
  char buff[REC_BUFF_SIZE];
  char *buff_idx;

  int
	num_records, /* Number of QSO records    */
	num_logged,  /* Number of QSL's printed  */
	fopen_ret;   /* Open_File() return value */

  if( isFlagClear(SAVE_RECORD) ) return( TRUE );
  ClearFlag( SAVE_RECORD );

  /* Create a file for QSO ADIF log if not open */
  strncpy( file_name, getenv("HOME"), 25 );
  strcat( file_name, "/xfhell/log.adif" );
  fopen_ret = Open_File( &rc_data.log_adif_fp, file_name, "a" );
  if( fopen_ret == FILE_NEW )
  {
	snprintf( buff, REC_BUFF_SIZE,
		"    QSO Log file created in ADIF v1.0 format"
		" by %s %s <EOH>\n\n", PACKAGE, VERSION );
	if( ! File_Print(&rc_data.log_adif_fp, buff) )
	  return( FALSE );
  }
  else
	if( fopen_ret == -1 )
	  return( FALSE );

  /* Print QSO record to adif format file */
  snprintf( buff, REC_BUFF_SIZE,
	  "<CALL:%d>%s<QSO_DATE:8>%s<TIME_ON:4>%s\n"
	  "<FREQ:%d>%s<MODE:4>Hell<RST_SENT:3>%3s<EOR>\n\n",
	  (int)strlen(qso_record.dx_call), qso_record.dx_call,
	  qso_record.date_adif, qso_record.time_adif,
	  (int)strlen(qso_record.freq), qso_record.freq,
	  qso_record.dx_rst );
  if( ! File_Print(&rc_data.log_adif_fp, buff) )
	return( FALSE );

  if( fflush(rc_data.log_adif_fp) != 0 )
  {
	perror( file_name );
	Error_Dialog(
		"Failed to fflush log.adif file\n"
		"Quit xfhell and correct", TRUE );
	return( FALSE );
  }

  /* Create a file for station log if not already open */
  strncpy( file_name, getenv("HOME"), 25 );
  strcat( file_name, "/xfhell/log.txt" );
  fopen_ret = Open_File( &rc_data.station_log_fp, file_name, "r+" );
  if( fopen_ret == FILE_NEW )
  {
	/* Print a header to log file */
	snprintf( buff, REC_BUFF_SIZE,
		"|------------------------------------"
		"--------------------------------------|\n"
		"|          Station Log File - Created in"
		" Text format by %6s-%3s         |\n"
		"|------------------------------------"
		"--------------------------------------|\n",
		PACKAGE, VERSION );
	if( ! File_Print(&rc_data.station_log_fp, buff) )
	  return( FALSE );

	/* Print an initial (0) count of records and qsl's */
	snprintf( buff, REC_BUFF_SIZE,
		"|  Num of QSO Records:%-6d  Records Logged:%-6d"
		"  Size of Records:%-4d  |\n"
		"|------------------------------------"
		"--------------------------------------|\n",
		0, 0, RECORD_SIZE );
	if( ! File_Print(&rc_data.station_log_fp, buff) )
	  return( FALSE );
  }
  else
	if( fopen_ret == -1 )
	  return( FALSE );

  /*** Increment and print Record and QSL count ***/
  /* Go to beginning of log file */
  fseek( rc_data.station_log_fp, 0, SEEK_SET );

  /* Find line with "Number of QSO Records:" */
  if( ! Locate_Line(&buff_idx, buff, "QSO Records:", rc_data.station_log_fp) )
	return( FALSE );
  num_records = atoi(buff_idx);

  /* Find "Number of QSO's QSL'd:" string */
  if( ! Locate_String(&buff_idx, buff, "Records Logged:") )
	return( FALSE );
  num_logged = atoi(buff_idx);

  /* Go back to beginning of line and print new record */
  fseek( rc_data.station_log_fp, -(1 + strlen(buff)), SEEK_CUR );
  snprintf( buff, REC_BUFF_SIZE,
	  "|  Num of QSO Records:%-6d  Records Logged:%-6d"
	  "  Size of Records:%-4d  |\n"
	  "|------------------------------------"
	  "--------------------------------------|\n",
	  ++num_records, num_logged, RECORD_SIZE );
  if( ! File_Print(&rc_data.station_log_fp, buff) )
	return( FALSE );

  /*** Print record in text log file ***/
  fseek( rc_data.station_log_fp, 0, SEEK_END );
  snprintf( buff, REC_BUFF_SIZE,
	  "|DX/ CALL: %-14s NAME: %-12s QTH: %-12s  LOC: %-6s|\n"
	  "|MY/ CALL: %-14s ZONE: %-12s QTH: %-12s  LOC: %-6s|\n"
	  "|QSO DATE: %-11s    TIME: %-5s UTC   FREQ: %-13s QSL: NO    |\n"
	  "|QSO MODE: %-11s DX-REPT: %-3s      MY-REPT: %-3s     VIA: %-12s|\n"
	  "|MY TRANS: %-11s   POWER: %-9s    ANT: %-15s          |\n"
	  "|MY RECVR: %-11s   N.FIG: %-7s      ANT: %-15s          |\n"
	  "| REMARKS: %-45s                   |\n"
	  "|------------------------------------"
	  "--------------------------------------|\n",
	  qso_record.dx_call, qso_record.dx_name, qso_record.dx_qth, qso_record.dx_loc,
	  rc_data.call, rc_data.zone, rc_data.qth, rc_data.loc,
	  qso_record.date, qso_record.time, qso_record.freq,
	  qso_record.mode, qso_record.dx_rst, qso_record.my_rst, qso_record.via,
	  rc_data.tx, rc_data.tx_power, rc_data.tx_ant,
	  rc_data.rx, rc_data.rx_nfig, rc_data.rx_ant, qso_record.remarks );
  if( ! File_Print(&rc_data.station_log_fp, buff) )
	return( FALSE );

  return( TRUE );

} /* Save_QSO_Record() */

/*------------------------------------------------------------------------*/

/*  Read_QSO_Record()
 *
 *  Reads and validates QSO record entries
 */

  gboolean
Read_QSO_Record( void )
{
  GtkWidget *entry;

  /* Enter field values to QSO record structure */
  entry = lookup_widget( main_window, "callsign" );
  strncpy( qso_record.dx_call,	gtk_entry_get_text(GTK_ENTRY(entry)), 15 );
  entry = lookup_widget( main_window, "rst_in" );
  strncpy( qso_record.my_rst,	gtk_entry_get_text(GTK_ENTRY(entry)),  4 );
  entry = lookup_widget( main_window, "rst_out" );
  strncpy( qso_record.dx_rst,	gtk_entry_get_text(GTK_ENTRY(entry)),  4 );
  entry = lookup_widget( main_window, "op_name" );
  strncpy( qso_record.dx_name,	gtk_entry_get_text(GTK_ENTRY(entry)), 13 );
  entry = lookup_widget( main_window, "qth_name" );
  strncpy( qso_record.dx_qth,	gtk_entry_get_text(GTK_ENTRY(entry)), 13 );
  entry = lookup_widget( main_window, "locator" );
  strncpy( qso_record.dx_loc,	gtk_entry_get_text(GTK_ENTRY(entry)),  7 );
  entry = lookup_widget( main_window, "band" );
  strncpy( qso_record.freq,	gtk_entry_get_text(GTK_ENTRY(entry)),  6 );

  /* Enter operating mode */
  strcpy( qso_record.mode, "Hell" );

  /* Validate QSO Record */
  if( (strlen(qso_record.dx_call) > 2) &&
	  (strlen(qso_record.my_rst) == 3) &&
	  (strlen(qso_record.freq)    > 0) )
	return( TRUE );
  else
	return( FALSE );

} /* Read_QSO_Record() */

/*------------------------------------------------------------------------*/

/*  Clear_Record_Fields()
 *
 *  Clears all QSO Record fields
 */

  void
Clear_Record_Fields( gboolean all )
{
  /* Save outstanding records */
  if( isFlagSet(SAVE_RECORD) )
	Save_QSO_Record();
  gtk_entry_set_text( GTK_ENTRY(lookup_widget(main_window, "callsign")), "" );
  gtk_entry_set_text( GTK_ENTRY(lookup_widget(main_window, "rst_in")), "" );
  gtk_entry_set_text( GTK_ENTRY(lookup_widget(main_window, "rst_out")), "" );
  gtk_entry_set_text( GTK_ENTRY(lookup_widget(main_window, "op_name")), "" );
  gtk_entry_set_text( GTK_ENTRY(lookup_widget(main_window, "qth_name")), "" );
  gtk_entry_set_text( GTK_ENTRY(lookup_widget(main_window, "locator")), "" );
  if( all )
	gtk_entry_set_text( GTK_ENTRY(lookup_widget(main_window, "band")), "" );

  qso_record.dx_call[0]    = '\0';
  qso_record.my_rst[0]  = '\0';
  qso_record.dx_rst[0] = '\0';
  strcpy( qso_record.dx_name, "OP" );
  qso_record.dx_qth[0]     = '\0';
  qso_record.dx_loc[0]     = '\0';
  if( all )
	qso_record.freq[0]  = '\0';

} /* Clear_Record_Fields() */

/*------------------------------------------------------------------------*/

/*  Open_Record_File()
 *
 *  Opens a file for raw QSO recording
 */

  gboolean
Open_Record_File(void)
{
  /* File path for QSO record  */
  char qso_record_fpath[64];


  /* Open file for QSO Record, abort on error */
  if( isFlagSet(RECORD_QSO) )
  {
	/* Create a file for recording QSO's */
	if( rc_data.qso_record_fp == NULL )
	{
	  /* Make file path for record file */
	  snprintf( qso_record_fpath, 64, "%s/xfhell/record.txt", getenv("HOME") );

	  /* Open file, abort on error */
	  rc_data.qso_record_fp = fopen( qso_record_fpath, "a" );
	  if( rc_data.qso_record_fp == NULL )
	  {
		perror( qso_record_fpath );
		Error_Dialog(
			"Failed to open QSO Record file\n"
			"Quit xfhell and correct", TRUE );
		return( FALSE );
	  }

	} /* if( QSO_record_fp == NULL ) */

  } /* if( isFlagSet(RECORD_QSO) ) */

  return( TRUE );

} /* Open_Record_File(void) */

/*------------------------------------------------------------------------*/

/*  Set_TxRx_Labels()
 *
 *  Sets up the labels in the Tx/Rx frames
 */

  void
Set_TxRx_Labels( void )
{
  char lbl[80];

  /* Set Tx button label */
  if( isFlagSet(TRANSMIT_KEYBD) ||
	  isFlagSet(TRANSMIT_MACRO) ||
	  isFlagSet(KEYBD_BUSY) )
  {
	gtk_label_set_markup( GTK_LABEL(lookup_widget(main_window, "xmit_status")),
		XMIT_ON );
	gtk_spin_button_set_value( GTK_SPIN_BUTTON(lookup_widget(
			main_window, "volume_spinbutton")), rc_data.vol_lev );
  }
  else
	gtk_label_set_markup( GTK_LABEL(lookup_widget(main_window, "xmit_status")),
		XMIT_OFF );

  /* Set Tx fonts label */
  gtk_entry_set_text( GTK_ENTRY(lookup_widget(main_window, "font_entry")),
	  rc_data.font_data.font_name );

  /* Set Tx/Rx mode and baud rate */
  if( isFlagSet(MODE_FELDHELL) )
	snprintf( lbl, 80, "FeldHell @ %1.1f Baud",
		rc_data.baud_rate * (double)rc_data.dot_resol );
  else
	if( isFlagClear(MODE_FELDHELL) )
	  snprintf( lbl, 80, "FMHell @ %1.1f Baud",
		  rc_data.baud_rate * (double)rc_data.dot_resol);
  gtk_entry_set_text( GTK_ENTRY(lookup_widget(main_window, "mode_entry")), lbl );

  /* Set Rx button label */
  if( isFlagSet(RECEIVE_MODE) )
  {
	if( isFlagSet(KEYBD_BUSY) )
	  gtk_label_set_markup( GTK_LABEL(lookup_widget(
			  main_window, "rcve_status")), RECV_SBY );
	else
	{
	  if( isFlagSet(AUTO_DESKEW) )
		snprintf( lbl, 80, RECV_ON, "Deskew" );
	  else
		snprintf( lbl, 80, RECV_ON, "Normal" );

	  gtk_label_set_markup( GTK_LABEL(lookup_widget(
			  main_window, "rcve_status")), lbl );
	  gtk_spin_button_set_value( GTK_SPIN_BUTTON(lookup_widget(
			  main_window, "volume_spinbutton")), rc_data.rec_lev );
	}
  }
  else
	gtk_label_set_markup( GTK_LABEL(lookup_widget(
			main_window, "rcve_status")), RECV_OFF );

} /* Set_TxRx_Labels() */

/*------------------------------------------------------------------------*/

/*  Load_Fonts()
 *
 *  Loads a .bdf font file into a font buffer
 */

  gboolean
Load_Fonts( char *font_file )
{
  FILE *font_fp = NULL;

  int
	bmx,
	bmy,
	dot_cnt,
	new_dot,
	last_dot,
	glyph_idx,
	bmap_idx,
	old_fbb_h,
	idx;

  /* Buffer for Load_Line  */
  char line[81];


  /* Open font file, abort on error */
  font_fp = fopen( font_file, "r" );
  if( font_fp == NULL )
  {
	perror( font_file );
	Error_Dialog( "Failed to open Font file", FALSE );
	return( FALSE );
  }

  /* Find font bounding box */
  if( ! Find_Line(line, "FONTBOUNDINGBOX ", font_fp) )
  {
	Error_Dialog(
		"Error reading Font file\n"
		"FONTBOUNDINGBOX entry not found", TRUE );
	return( FALSE );
  }

  idx = 14;
  Find_String( line, &idx );
  rc_data.font_data.fbb_w = atoi( &line[idx] );
  Find_String( line, &idx );
  old_fbb_h = rc_data.font_data.fbb_h;
  rc_data.font_data.fbb_h = atoi( &line[idx] );
  Find_String( line, &idx );
  rc_data.font_data.fbb_x = atoi( &line[idx] );
  Find_String( line, &idx );
  rc_data.font_data.fbb_y = atoi( &line[idx] );

  /* If fbbx height changes, clear Rx window */
  if( old_fbb_h != rc_data.font_data.fbb_h )
	SetFlag( CLEAR_RX_WINDOW );

  /* Check fbbx size and position */
  if( (rc_data.font_data.fbb_w > 30) || (rc_data.font_data.fbb_w <  5) ||
	  (rc_data.font_data.fbb_x != 0) || (rc_data.font_data.fbb_y < -4) )
  {
	Error_Dialog(
		"Error reading Font file\n"
		"FONTBOUNDINGBOX data incorrect", TRUE );
	return( FALSE );
  }

  /* bitmap_hgt & bitmap_wdt are the height and width in pixels of
   * bitmap of Hell characters. This is the product of dot size in
   * pixels * size in dots of Hell characters. Basic Hellschreiber
   * characters are 7x7 dots but vertically dots are two pixels
   * high so the Tx font height is 2x the character size in dots.
   */
  rc_data.bitmap_hgt = rc_data.dot_hgt * rc_data.font_data.fbb_h / 2;
  rc_data.bitmap_wdt = rc_data.dot_wdt * rc_data.font_data.fbb_h / 2;

  /* Read font name */
  if( ! Find_Line(line, "FONT_NAME ", font_fp) )
  {
	Error_Dialog(
		"Error reading Font file\n"
		"FONT_NAME entry not found", FALSE );
	return( FALSE );
  }

  /* Remove spaces */
  idx = 8;
  Find_String( line, &idx );
  strcpy( rc_data.font_data.font_name, &line[idx] );

  /* Read number of characters */
  if( ! Find_Line(line, "CHARS ", font_fp) )
  {
	Error_Dialog(
		"Error reading Font file\n"
		"CHARS entry not found", FALSE );
	return( FALSE );
  }
  rc_data.font_data.num_glyphs = atoi( &line[6] );

  /* Allocate memory to glyph buffer */
  rc_data.font_data.glyph_data = malloc(
	  rc_data.font_data.num_glyphs * sizeof(glyph_data_t) );
  if( rc_data.font_data.glyph_data == NULL )
  {
	perror( "Font Buffer" );
	Error_Dialog(
		"malloc() for Font data buffer\n"
		"failed! Quit xfhell and correct", TRUE );
	return( FALSE );
  }

  /* Read all glyph data */
  for( glyph_idx = 0; glyph_idx < rc_data.font_data.num_glyphs; glyph_idx++ )
  {
	/* Read ENCODING */
	if( ! Find_Line(line, "STARTCHAR ", font_fp) )
	{
	  Error_Dialog(
		  "Error reading Font file\n"
		  "STARTCHAR entry not found", TRUE );
	  return( FALSE );
	}

	if( ! Find_Line(line, "ENCODING ", font_fp) )
	{
	  Error_Dialog(
		  "Error reading Font file\n"
		  "ENCODING entry not found", TRUE );
	  return( FALSE );
	}
	rc_data.font_data.glyph_data[glyph_idx].encoding = atoi( &line[9] );

	/* Read DWIDTH */
	if( ! Find_Line(line, "DWIDTH ", font_fp) )
	{
	  Error_Dialog(
		  "Error reading Font file\n"
		  "DWIDTH entry not found", TRUE );
	  return( FALSE );
	}

	idx = 5;
	Find_String( line, &idx );
	rc_data.font_data.glyph_data[glyph_idx].dwidth_x = atoi( &line[idx] )+1;
	Find_String( line, &idx );
	rc_data.font_data.glyph_data[glyph_idx].dwidth_y = atoi( &line[idx] );

	/* Read BBX */
	if( ! Find_Line(line, "BBX ", font_fp) )
	{
	  Error_Dialog(
		  "Error reading Font file\n"
		  "BBX entry not found", TRUE );
	  return( FALSE );
	}

	idx = 2;
	Find_String( line, &idx );
	rc_data.font_data.glyph_data[glyph_idx].bbx_w = atoi( &line[idx] );
	Find_String( line, &idx );
	rc_data.font_data.glyph_data[glyph_idx].bbx_h = atoi( &line[idx] );
	Find_String( line, &idx );
	rc_data.font_data.glyph_data[glyph_idx].bbx_x = atoi( &line[idx] );
	Find_String( line, &idx );
	rc_data.font_data.glyph_data[glyph_idx].bbx_y = atoi( &line[idx] );

	/* Check DWIDTH and BBX w are same. DWIDTH
	 * was incremented by 1 to leave char space */
	if( (rc_data.font_data.glyph_data[glyph_idx].bbx_w !=
		  rc_data.font_data.glyph_data[glyph_idx].dwidth_x-1) &&
		(rc_data.font_data.glyph_data[glyph_idx].encoding != 0x20) )
	  fprintf( stderr, "Encoding %2x: DWIDTH != BBX width\n",
		  rc_data.font_data.glyph_data[glyph_idx].encoding );

	/* Make sure BBX is at the x origin */
	if( rc_data.font_data.glyph_data[glyph_idx].bbx_x != 0 )
	  fprintf( stderr, "Encoding %2x: BBX x != 0\n",
		  rc_data.font_data.glyph_data[glyph_idx].encoding );

	/* Read and convert bitmap entries */
	bmap_idx =
	  rc_data.font_data.fbb_h -
	  rc_data.font_data.glyph_data[glyph_idx].bbx_h -
	  rc_data.font_data.glyph_data[glyph_idx].bbx_y +
	  rc_data.font_data.fbb_y;

	/* Allocate bitmap buffer */
	rc_data.font_data.glyph_data[glyph_idx].bitmap =
	  malloc( sizeof(int) * (size_t)rc_data.font_data.fbb_h );
	if( rc_data.font_data.glyph_data[glyph_idx].bitmap == NULL )
	{
	  perror( "Font bitmap buffer" );
	  Error_Dialog(
		  "malloc() for Font bitmap buffer\n"
		  "failed! Quit xfhell and correct", TRUE );
	  return( FALSE );
	}

	/* Clear unused part of bitmap */
	for( idx = 0; idx < bmap_idx; idx++ )
	  rc_data.font_data.glyph_data[glyph_idx].bitmap[idx] = 0;

	if( ! Find_Line(line, "BITMAP", font_fp) )
	{
	  Error_Dialog(
		  "Error reading Font file\n"
		  "BITMAP entry not found", FALSE );
	  return( FALSE );
	}

	/* Read hex bitmap and convert to binary bitmap */
	for( idx = 0; idx < rc_data.font_data.glyph_data[glyph_idx].bbx_h; idx++ )
	{
	  if( Load_Line(line, font_fp) == -1 )
		return( FALSE );
	  rc_data.font_data.glyph_data[glyph_idx].bitmap[bmap_idx++] = Hex2Bitmap(line);
	}

	/* Clear unused part of bitmap */
	for( idx = bmap_idx; idx < rc_data.font_data.fbb_h; idx++ )
	  rc_data.font_data.glyph_data[glyph_idx].bitmap[idx] = 0;

	/* Check that there are no lone 1 or 0 bits in glyph columns */
	dot_cnt = 1; last_dot = 0;
	for( bmx = 0; bmx < rc_data.font_data.glyph_data[glyph_idx].dwidth_x; bmx++ )
	{
	  /* Go up in bitmap buffer row by row */
	  for( bmy = rc_data.font_data.fbb_h-1; bmy >= 0; bmy-- )
	  {
		dot_cnt++;
		new_dot = rc_data.font_data.glyph_data[glyph_idx].bitmap[bmy] & (1 << (31-bmx) );
		if( new_dot != last_dot )
		{
		  /* Check for lone 1|0 bits */
		  if( dot_cnt < 2 )
			fprintf( stderr,
				"Encoding %2x: Single half-pixel in column %d\n",
				rc_data.font_data.glyph_data[glyph_idx].encoding, bmx );

		  /* Check for odd num. of half-pixels in Lo-res fonts */
		  if( (dot_cnt & 1) && (strstr(rc_data.font_data.font_name, "Lo") != NULL) )
		  {
			fprintf( stderr,
				"Encoding %2x: Odd number of half-pixels in column %d\n",
				rc_data.font_data.glyph_data[glyph_idx].encoding, bmx );
		  }

		  dot_cnt = 0;
		}
		last_dot = new_dot;

	  } /* for( bmy = rc_data.font_data.fbb_h-1; bmy >= 0; bmy-- ) */

	} /* for( bmx = 0; bmx < rc_data.font_data.glyph_data[glyph_idx].dwidth_x; bmx++ ) */

  } /* for( glyph_idx = 0; glyph_idx < font_data.num_glyphs; glyph_idx++ ) */

  /* Encoding num of first & last char in file */
  rc_data.font_data.first_glyph = rc_data.font_data.glyph_data[0].encoding;
  rc_data.font_data.last_glyph  =
	rc_data.font_data.first_glyph + rc_data.font_data.num_glyphs;

  fclose( font_fp );

  return( Alloc_Tx_Buff() );

} /* Load_Fonts() */

/*------------------------------------------------------------------------*/

/*  Find_Line()
 *
 *  Finds a line starting with a given string
 */

  gboolean
Find_Line( char *line, const char *label, FILE *fp )
{
  do
	if( Load_Line(line, fp) == -1 )
	{
	  Error_Dialog(
		  "Error reading Font file\n"
		  "Load_Line() has failed", FALSE );
	  return( FALSE );
	}
  while( strncmp( line, label, strlen(label) ) != 0 );

  return( TRUE );

} /* Find_Line() */

/*------------------------------------------------------------------------*/

/*  Find_String()
 *
 *  Finds a string following spaces
 */

  void
Find_String( char *line, int *idx )
{
  /* Look for a space */
  while( line[++*idx] != ' ' )
	if( (line[*idx] == CR) || (line[*idx] == LF) )
	{
	  Error_Dialog(
		  "Error reading Font file\n"
		  "Unexpected End-Of-Line", FALSE );
	  return;
	}

  /* Look for next non-space char */
  while( line[++*idx] == ' ' );
  if( (line[*idx] == CR) || (line[*idx] == LF) )
  {
	Error_Dialog(
		"Error reading Font file\n"
		"Unexpected End-Of-Line", FALSE );
	return;
  }
} /*  Find_String() */

/*------------------------------------------------------------------------*/

/*  Hex2Bitmap()
 *
 *  Convert a hex number to bitmap
 */

  unsigned int
Hex2Bitmap( char *hex_str )
{
  unsigned int bitmap;
  char hex[] = {
	'0','1','2','3','4','5','6','7',
	'8','9','A','B','C','D','E','F' };

  int
	str_idx,
	hex_idx;


  /* Convert characters of hex string to binary bitmap */
  str_idx = bitmap = 0;
  while( (str_idx < 8) && (hex_str[str_idx] != '\0') )
  {
	bitmap <<= 4;
	for( hex_idx = 0; hex_idx < 16; hex_idx++ )
	  if( hex[hex_idx] == hex_str[str_idx] )
		break;
	str_idx++;

	bitmap |= hex_idx;
  }

  /* Shift bitmap to uppermost bit */
  while( str_idx++ < 8 )
	bitmap <<= 4;

  return( bitmap );

} /* Hex2Bitmap( char *hex_str ) */

/*------------------------------------------------------------------------*/

/*  Render_Font()
 *
 *  Renders a font buffer on the Rx screen
 */

  gboolean
Render_Font( font_data_t *font, int glx )
{
  unsigned char
	dot_val, /* Dot value (black|white) */
	column[2*rc_data.font_data.fbb_h]; /* Column of dots */

  int
	bmx,   /* Bitmap x index */
	bmy,   /* Bitmap y index */
	fpix;  /* Pixels in font */


  /* Double size fonts selected */
  fpix = 2*rc_data.font_data.fbb_h-1;

  /* Use columns according to DWIDTH y of font */
  for( bmx = 0; bmx < font->glyph_data[glx].dwidth_x; bmx++ )
  {
	/* Go down in bitmap buffer row by row */
	for( bmy = 0; bmy < rc_data.font_data.fbb_h; bmy++ )
	{
	  /* Set black or white according to bit value */
	  if( font->glyph_data[glx].bitmap[bmy] & (1 << (31-bmx)) )
		dot_val = 0xff; /* Black dot (intensity is inverted) */
	  else
		dot_val = 0x00; /* White dot */

	  /* Two pixels per font's dot, vertically */
	  column[fpix   -2*bmy] = dot_val;
	  column[fpix-1 -2*bmy] = dot_val;
	}

	if( ! Draw_Column(column) )
	  return( FALSE );

  } /* for( bmx = font->glyph_data[glx].dwidth_y; bmx >= 0; bmx-- ) */

  return( TRUE );

} /* Render_Font() */

/*------------------------------------------------------------------------*/

/*  Save_Pixbuf()
 *
 *  Saves the Receive pixbuf as a jpeg image
 */

  gboolean
Save_Pixbuf( void )
{
  /* For saving pixbuf */
  char pixbuf_fpath[80];
  GError *error = NULL;
  gboolean err;
  int len;

  /* Variables for reading time (UTC) */
  time_t tp;
  struct tm utc;

  if( isFlagClear(SAVE_PIXBUF) ) return(TRUE);
  ClearFlag( SAVE_PIXBUF );

  /* Make a file path for pixbuf file */
  snprintf( pixbuf_fpath, 80, "%s/xfhell/pixbufs/", getenv("HOME") );

  /* Prepare a file name as UTC date-time. */
  /* Default paths are images/ and record/ */
  time( &tp );
  utc = *gmtime( &tp );
  len = strlen( pixbuf_fpath );
  strftime( &pixbuf_fpath[len], 21, "%d%b%Y-%H%M%S.jpg", &utc );

  /* Save Receive pixbuf */
  err = gdk_pixbuf_save( gbl_pixbuf, pixbuf_fpath,
	  "jpeg", &error, "quality", "75", NULL );
  if( error != NULL )
  {
	fprintf( stderr, error->message );
	Error_Dialog( "Failed to save Receive pixbuf\n", FALSE );
  }

  return( err );

} /* Save_Pixbuf() */

/*------------------------------------------------------------------------*/

/* Alloc_Tx_Buff()
 *
 * Allocates memory to transmit samples buffer
 */

  gboolean
Alloc_Tx_Buff(void)
{
  int idx;

  rc_data.txbuff_len =
	rc_data.font_data.fbb_h *
	rc_data.samples_perdot2 *
	rc_data.num_chan;

  /* Allocate memory to samples buffer */
  dsp_buff = realloc( dsp_buff, rc_data.txbuff_len );
  if( dsp_buff == NULL )
  {
	perror( "DSP buffer" );
	Error_Dialog(
		"DSP buffer realloc() failed\n"
		"Quit xfhell and correct", TRUE  );
	return( FALSE );
  }

  /* Clear buffer */
  for( idx = 0; idx < rc_data.txbuff_len; idx++ )
	dsp_buff[idx] = 0x80;

  return( TRUE );

} /* Alloc_Tx_Buff() */

/*------------------------------------------------------------------------*/

/*  Process_RST()
 *
 *  Processes RST-out and RST-in fields
 */

  void
Process_RST( GtkEditable *editable )
{
  G_CONST_RETURN gchar *rst;
  int len, idx;

  /* Validate RST: numerics only, */
  /* min 337 max 559, ignore blank field */
  rst = gtk_entry_get_text( GTK_ENTRY(editable) );
  len = strlen(rst);

  /* Ignore blank fields */
  if( len == 0 )
	return;

  for( idx = 0; idx < len; idx++ )
  {
	/* Reject non-numbers */
	if( (rst[idx] < '0') ||
		(rst[idx] > '9') )
	{
	  Bad_Entry_Dialog(
		  "Invalid character entered\n"
		  "Numbers only allowed" );
	  return;
	}

	/* Reject RST < 337 or > 599 */
	switch( idx )
	{
	  case 0:
		if( (rst[idx] < '3') ||
			(rst[idx] > '5') )
		{
		  Bad_Entry_Dialog(
			  "Number out of range\n"
			  "Min = 3 and Max = 5" );
		  return;
		}
		break;

	  case 1:
		if( rst[idx] < '3' )
		{
		  Bad_Entry_Dialog(
			  "Number out of range\n"
			  "Min = 3 and Max = 9" );
		  return;
		}
		break;

	  case 2:
		if( rst[idx] < '7' )
		{
		  Bad_Entry_Dialog(
			  "Number out of range\n"
			  "Min = 7 and Max = 9" );
		  return;
		}
	} /* switch( idx ) */

  } /* for( idx = 0; idx < len; idx++ ) */

} /* Process_RST() */

/*------------------------------------------------------------------------*/

/*  Select_Macro()
 *
 *  Selects a macro to be transmitted
 */

  void
Select_Macro( macro_num )
{
  /* Transmit a macro */
  if( isFlagClear(TRANSMIT_MACRO) )
  {
	SetFlag( TRANSMIT_MACRO );
	ClearFlag( RECEIVE_MODE );
	gbl_macro_idx = macro_num - 1;
	Set_TxRx_Labels();
	gtk_idle_add( Transmit_Macro, NULL );
  }
  else
	ClearFlag( TRANSMIT_MACRO );

} /* Select_Macro() */

/*------------------------------------------------------------------------*/

/*  Open_File()
 *
 *  Opens a file
 */

  int
Open_File( FILE **fp, char *fpath, const char *mode )
{
  struct stat buf;

  /* Check if file exixts */
  if( *fp == NULL )
  {
	/* If file does not exist, open new */
	if( stat(fpath, &buf) == -1 )
	{
	  /* Create new file */
	  if( (*fp = fopen(fpath, "w+")) == NULL )
	  {
		char mesg[110];
		snprintf( mesg, 110, 
			"Failed to create\n%s\n"
			"Quit xfhell and correct", fpath );
		perror( fpath );
		Error_Dialog( mesg, TRUE );
		return( -1 );
	  }
	  else
	  {
		/* Re-open in requested mode */
		Close_File( fp );
		if( (*fp = fopen(fpath, mode)) == NULL )
		{
		  char mesg[110];
		  snprintf( mesg, 110,
			  "Failed to open\n%s\n"
			  "Quit xfhell and correct", fpath );
		  perror( fpath );
		  Error_Dialog( mesg, TRUE );
		  return( -1 );
		}
		else
		  return( FILE_NEW );
	  }
	} /* if( stat(fpath, buf) == -1 ) */
	else /* If file exists, re-open in required mode */
	{
	  if( (*fp = fopen(fpath, mode)) == NULL )
	  {
		char mesg[110];
		snprintf( mesg, 110,
			"Failed to open\n%s\n"
			"Quit xfhell and correct", fpath );
		perror( fpath );
		Error_Dialog( mesg, TRUE );
		return( -1 );
	  }
	  else
		return( FILE_OPEN );

	} /* else */

  } /* if( *fp == NULL ) */

  return( FILE_OPEN );

} /* Open_File() */

/*------------------------------------------------------------------------*/

/*  Close_File()
 *
 *  Closes and NULL's an open file pointer
 */

  void
Close_File( FILE **fp )
{
  if( *fp != NULL )
  {
	fclose( *fp );
	*fp = NULL;
  }
}

/*------------------------------------------------------------------------*/

/*  File_Print()
 *
 *  Prints a string to a file
 */

  gboolean
File_Print( FILE **fp, char *string )
{
  if( fprintf(*fp, "%s", string) != strlen(string) )
  {
	perror( "xfhell: fprintf" );
	Error_Dialog(
		"Error printing to file\n"
		"Quit xfhell and correct", TRUE );
	Close_File( fp );
	return( FALSE );
  }

  return( TRUE );

} /* File_Print() */

/*------------------------------------------------------------------------*/

/*  Find_Line()
 *
 *  Finds a line in a text file containing a given string
 */

  gboolean
Locate_Line( char **line_idx, char *line, const char *string, FILE *fp )
{
  int err;

  /* Load lines and look for string */
  do
  {
	err = Load_Line( line, fp );
	*line_idx = strstr(line, string);
  }
  while( (err == 0) && (*line_idx == NULL) );

  /* Usually EOF is the cause of error */
  if( err )
  {
	Error_Dialog(
		"Error searching text file\n"
		"May be empty or corrupted", FALSE );
	return( FALSE );
  }

  *line_idx += strlen(string);

  return( TRUE );

} /* Find_Line() */

/*------------------------------------------------------------------------*/

/*  Locate_String()
 *
 *  Finds a string in a buffer
 */

  gboolean
Locate_String( char **line_idx, char *line, const char *string )
{
  if( (*line_idx = strstr(line, string)) == NULL )
  {
	Error_Dialog(
		"Error searching station_log.txt\n"
		"May be empty or corrupted", TRUE );
	return( FALSE );
  }

  *line_idx += strlen(string);

  return( TRUE );

} /* Locate_String() */

/*------------------------------------------------------------------------*/

/* New_Parameters()
 *
 * Sets up parameters in the program after one changes
 */

  void
New_Parameters( void )
{
  int idx;
  char btn[6], font_file[81];
  double baud_rate[BAUD_RATES] =
  { BAUD245, BAUD210, BAUD122, BAUD105, BAUD61, BAUD30, BAUD15, BAUD7 };

  if( isFlagSet(NO_CALLBACKS) ) return;

  /* Find active baud rate menu item */
  for( idx = 0; idx < BAUD_RATES; idx++ )
  {
	snprintf( btn, 6, "br%d", (int)baud_rate[idx] );
	if( gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(
			lookup_widget(gbl_popup_menu, btn))) )
	  break;
  }
  rc_data.baud_rate = baud_rate[idx];
  rc_data.dot_resol = 1;

  /* Dot resolution=2 for FMHell Hi-Res transmissions */
  if( rc_data.baud_rate == BAUD245 )
  {
	rc_data.baud_rate = BAUD122;
	rc_data.dot_resol = 2;
  }
  else
	if( rc_data.baud_rate == BAUD210 )
	{
	  rc_data.baud_rate = BAUD105;
	  rc_data.dot_resol = 2;
	}

  /* Prepare default Font file path if needed */
  if( strlen(rc_data.font_file) )
  {
	strncpy( font_file, rc_data.font_file, 80 );
	rc_data.font_file[0] = '\0';
  }
  else
  {
	/* Prepare font path */
	strncpy( font_file, getenv("HOME"), 38 );
	strcat(  font_file, "/xfhell/fonts/" );
	if( rc_data.baud_rate == BAUD105 )
	{
	  strcat( font_file, "12pt/" );
	  if( isFlagClear(MODE_FELDHELL) && (baud_rate[idx] == BAUD105) )
		strncat( font_file, FONT_LOWRES, 20 );
	  else
		strncat( font_file, FONT_NRMRES, 20 );
	}
	else
	{
	  strcat( font_file, "14pt/" );
	  strncat( font_file, FONT_NRMRES, 20 );
	}
  }
  font_file[80] = '\0';
  /* Free font buffers before loading new fonts */
  for( idx = 0; idx < rc_data.font_data.num_glyphs; idx++ )
	free( rc_data.font_data.glyph_data[idx].bitmap );
  free( rc_data.font_data.glyph_data );
  Load_Fonts( font_file );

  /* Duration of various elements in DSP samples */
  rc_data.samples_perdot   = (int)((double)rc_data.dsp_speed/rc_data.baud_rate);
  /* Round to nearest int divisible by 4 */
  idx = rc_data.samples_perdot % 4;
  if( idx < 2 )
	rc_data.samples_perdot -= idx;
  else
	  rc_data.samples_perdot += 4-idx;
  rc_data.samples_perdot2  = rc_data.samples_perdot / 2;
  rc_data.samples_perpixel = rc_data.samples_perdot / rc_data.dot_hgt;
  rc_data.samples_percycle = rc_data.dsp_speed / rc_data.tone_freq;
  rc_data.fft_stride       = rc_data.samples_percycle / 4;

  /* Height and width in pixels of bitmap of Hell characters */
  rc_data.bitmap_hgt = (rc_data.dot_hgt * rc_data.font_data.fbb_h) / 2;
  rc_data.bitmap_wdt = ((rc_data.dot_hgt-1) * rc_data.font_data.fbb_h) / 2;

  /* Make wavetables for FeldHell and CW id's */
  Make_Cos_Wavetable();
  Make_Sin_Wavetable();

  SetFlag( NEW_BAUD_RATE );
  SetFlag( CLEAR_RX_WINDOW );
  Set_TxRx_Labels();
  Alloc_Tx_Buff();

  return;

} /* New_Parameters() */

/*------------------------------------------------------------------------*/

