/*  xfhell.h
 *
 *  Header file for xfhell
 */

/*
 *  xfhell: An application to transmit and receive
 *  Feld Hell signals using a computer's sound card
 *
 *  Copyright (C) 2001  Neoklis Kyriazis
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */


#ifndef XFHELL_H
#define XFHELL_H 1

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <signal.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <stdio.h>
#include <termios.h>
#include <string.h>
#include <time.h>
#include <math.h>

/* General definitions for signal processing */
#define RX_BUFFER_SIZE		1024 /* Receive capture buffer size */
#define KEY_BUFF_SIZE		256  /* Keyboard buffer size */

/* Signal Detector parameter definitions */
#define RESAMPLE_CONST		20000000	/* Used to calc. resample frequency */
#define MAXSEG_AVE_WIN		5			/* Max seg ave window len (samples)*/
#define ELEMENT_MAX			100			/* Defined max val of dot elements */
#define ROW_AVE_WIN			20			/* Length of row averaging window  */
#define BINMAX_AVE_WIN		5			/* Length of bim max averaging window */
#define FFT_SIZE			512			/* Size of fft input/output arrays  */
#define WFALL_WIDTH			256			/* Num. of waterfall display columns */
#define GOERTZEL_SCALE		2048		/* Scale factor for integer Goertzel algorithm */
#define SQUELCH_WINDOW		10			/* Length of squelch averager window */
#define PHASE_AVE_WINDOW	5			/* Length of phase averager window */
#define MAGNIPHASE_SCALE	44			/* Receive magniphase scale factor */

/* Baud rate definitions. They are slightly different
 * from nominal to make integer arithmetic more accurate */
#define BAUD245		245.0
#define BAUD210		210.0
#define BAUD122		122.5
#define BAUD105		105.0
#define BAUD61		 61.25
#define BAUD30		 30.625
#define BAUD15		 15.3125
#define BAUD7		  7.65625
#define BAUD_RATES	  8

/* Default normal and low-res (FM105) font file */
#define FONT_NRMRES	"FeldFatEn.bdf"
#define FONT_LOWRES	"FMFatLoEn.bdf"

/* Flag definitions */
#define DSP_IO_ERROR	0x00000001
#define ENABLE_CAT		0x00000002
#define RECORD_QSO		0x00000004
#define CAPITALIZE		0x00000008
#define SOUND_SETUP		0x00000010
#define TRANSMIT_MODE	0x00000020
#define RECEIVE_MODE	0x00000040
#define TRANSMIT_MACRO	0x00000080
#define TRANSMIT_TAG	0x00000100
#define TRANSMIT_KEYBD	0x00000200
#define SAVE_RECORD		0x00000400
#define SERIAL_IO_ERROR	0x00000800
#define ENABLE_SCOPE	0x00001000
#define CLEAR_TX_WINDOW	0x00002000
#define CLEAR_RX_WINDOW	0x00004000
#define AUTO_DESKEW		0x00008000
#define TX2RX_LOOPBACK	0x00010000
#define KEYBD_BUSY		0x00020000
#define SAVE_PIXBUF		0x00040000
#define ENABLE_CAT_847	0x00080000
#define ENABLE_CAT_857	0x00100000
#define NEW_BAUD_RATE	0x00200000
#define REVERSE_VIDEO	0x00400000
#define MODE_FELDHELL	0x00800000
#define CHAN_A_SELECT	0x01000000
#define NO_CALLBACKS	0x02000000

/* Tone control flags */
#define RAISE_TONE		0x01
#define TONE_RAISED		0x02
#define LOWER_TONE		0x02
#define TONE_LOWERED	0x01

/* Number and size of labels for commands  */
#define NUM_OF_LABELS		9
#define LABEL_LENGTH		17

/* Max length of line in text files */
#define LINE_LENGTH		77

/* Some character definitions */
#define CR			0x0d /* Carriage Return */
#define LF			0x0a /* Line Feed       */
#define BS			0x08 /* Back Space      */
#define IDLE_CHAR	0x1f /* Idle character  */

/* Number of recognizable tags */
#define NUM_OF_TAGS	13
/* Table of tag strings */
#define TAG_TABLE \
{ \
  "own-call", "own-name" , "own-qth" ,\
  "own-loc" , "own-rst"  , "rem-call",\
  "rem-name", "rem-qth"  , "rem-loc" ,\
  "rem-rst" , "date-time", "op-freq" ,\
  "app-version" \
}

/* Receive and Transmit status indicators */
#define RECV_OFF	"<span background=\"grey\" > RECEIVE (OFF) </span>"
#define RECV_ON		"<span background=\"green\" > RECEIVE (%s) </span>"
#define RECV_SBY	"<span background=\"orange\" > RECEIVE (Standby) </span>"
#define XMIT_OFF	"<span background=\"grey\" > TRANSMIT (OFF) </span>"
#define XMIT_ON		"<span background=\"red\" > TRANSMIT (ON) </span>"

/*-------------------------------------------------------------------*/

/* Definitions for ascii to Morse code conversion */
#define NUM_MORSE_CHARS    55 /* Number of chars in Morse table */

/* ASCII equivalents to Morse hex code. Last */
/* one ( ) used for unrecognized characters. */
#define ASCII_CHAR { \
  'A','B','C','D','E','F','G','H','I','J','K','L','M', \
  'N','O','P','Q','R','S','T','U','V','W','X','Y','Z', \
  '1','2','3','4','5','6','7','8','9','0','.',',',':', \
  '?','\'','-','/','(','"',';','$','#','<','!','@',']',\
  '=','~',' ' }

/* Hex equivalents to Morse code chars above. Formed   */
/* by starting with a 1 and following with a 0 for     */
/* dash and 1 for dit e.g: A = ditdah = 101 = Hex 0x06 */
#define MORSE_CODE { \
  0x06,0x17,0x15,0x0b,0x03,0x1d,0x09,0x1f,0x07,0x18,0x0a,0x1b,0x04,\
  0x05,0x08,0x19,0x12,0x0d,0x0f,0x02,0x0e,0x1e,0x0c,0x16,0x14,0x13,\
  0x30,0x38,0x3c,0x3e,0x3f,0x2f,0x27,0x23,0x21,0x20,0x6a,0x4c,0x47,\
  0x73,0x61,0x5e,0x2d,0x52,0x6d,0x55,0xf6,0x35,0x7a,0x2a,0x37,0x29,\
  0x2e,0xff,0x01 }

/* Arc tan correction table */
#define ARC_TAN_TABLE \
{ \
   0,  1,  3,  4,  5,  6,  8,  9, 10, 11, 13, 14, \
  15, 16, 17, 18, 20, 21, 22, 23, 24, 25, 26, 27, \
  28, 29, 30, 31, 32, 33, 34, 35, 35, 36, 37, 38, \
  39, 39, 40, 41, 42, 42, 43, 44, 44, 45          \
}

/*-------------------------------------------------------------------*/

/* Glyph data structure */
typedef struct
{
  int
	encoding,
	dwidth_x,
	dwidth_y,
	bbx_w,
	bbx_h,
	bbx_x,
	bbx_y;

  unsigned int *bitmap;

} glyph_data_t;

/* Font data structure */
typedef struct
{
  /* Name of current font */
  char font_name[81];

  /* Font bounding box */
  int fbb_w, fbb_h, fbb_x, fbb_y;

  int
	first_glyph, /* First character in font file */
	last_glyph,  /* Last character in font file  */
	num_glyphs;  /* Number of chars in font file */

  glyph_data_t *glyph_data;

} font_data_t;

/* Configuration parameters structure */
typedef struct
{
  char
	mixer[15],   /* Mixer device     */
	dsp_in[15],  /* Input DSP device */
	dsp_out[15], /* Outpt DSP device */
	rec_src[10], /* Capture source   */
	lev_dev[10], /* Rec level device */
	mode[7],     /* Mono | Stereo    */
	channel[6],  /* Left | Right     */
	serial1[15], /* CAT serial port 1 */
	serial2[15], /* CAT serial port 2 */
	cw_msg[25];  /* CW message to send*/

  gchar font_file[81]; /* Font file name */

  double
	baud_rate; /* Hellschreiber baud rate */

  int
	rec_lev,		/* Capture/record level	*/
	vol_lev,		/* Output volume level	*/
	wpm,			/* Morse encoder speed wpm */
	wwrap,			/* Word wrap column number */
	dot_hgt,		/* Receive dot height in pixels */
	dot_wdt,		/* Receive dot width in pixels  */
	bitmap_hgt,		/* Receive bitmap height in pixels */
	bitmap_wdt,		/* Receive bitmap width in pixels  */
	num_rows,		/* Number of character rows in Rx window */
	num_chan,		/* Stereo (2) or Mono(1) selection */
	chan_used,		/* Channel in use: left=0, right=1 */
	samples_percycle, /* Samples per cycle of audio freq. */
	samples_perpixel, /* Samples per screen pixel */
	samples_perdot,   /* Samples per Hell character dot */
	samples_perdot2,  /* Samples per Hell character half dot */
	dot_resol,		  /* =2 for FMHell Hi-Res fonts, else =1 */
	dsp_speed,		/* Sound card DSP speed samples/sec */
	tone_freq, 		/* Input tone (Rx BFO) frequency */
	fft_stride,		/* FFT's stride over input data  */
	red_line,		/* Position of tuning reference in fft window */
	txbuff_len,		/* Tx signal samples buffer size */
	rxbuff_len,		/* Rx signal samples buffer size */
	dsp_in_fd,		/* Capture dsp fd  */
	dsp_out_fd,		/* Playback dsp fd */
	mix_fd;			/* Mixer device fd */

  FILE
	*qso_record_fp,  /* File for recording QSO's */
	*log_adif_fp,    /* File for ADIF QSO record */
	*station_log_fp; /* File for station's log   */

  /* Local operator data */
  char
	call[15],
	name[13],
	zone[15],
	qth[13],
	loc[7];

  /* Tx/Rx power, N.Fig, antennas */
  char
	tx[12],
	tx_power[10],
	tx_ant[16],
	rx[12],
	rx_nfig[8],
	rx_ant[16];

  font_data_t font_data; /* Font data buffer */

} rc_data_t;

/* QSO record type definition */
typedef struct
{
  /* QSO data */
  char
	dx_call[15],  /* Remote (dx) station callsign */
	dx_rst[4],    /* My RST report to Dx station  */
	my_rst[4],    /* My RST report to Dx station  */
	dx_name[13],  /* Remote station operator name */
	dx_qth[13],   /* Remote station QTH loc. name */
	dx_loc[7],    /* Remote station QTH locator   */
	date[12],     /* Normal format QSO date  */
	date_adif[9], /* QSO date in ADIF format */
	time[6],      /* Normal format QSO time  */
	time_adif[5], /* QSO time in ADIF format */
	freq[14],     /* QSO operating frequency MHz */
	mode[12],     /* Mode of operation */
	via[13];      /* Via which path or relay Sat */

  /* Remarks on QSO */
  char remarks[46];

} qso_record_t;

/*-------------------------------------------------------------------*/

/* Function prototypes produced by cproto */
/* callbacks.c */
void Error_Dialog(char *message, gboolean hide);
void Quit_Dialog(char *message);
void Save_Dialog(char *message);
void Bad_Entry_Dialog(char *message);
void on_main_window_destroy(GtkObject *object, gpointer user_data);
/* cat.c */
void Open_Tcvr_Serial(void);
void Read_Tcvr_Serial(unsigned char *cmnd, int block_size);
void Read_Rx_Status(char *mode, char *freq, char *rst);
void Write_Tcvr_Command(unsigned char *cmnd);
void Set_Tcvr_Mode(void);
void Close_Tcvr_Serial(void);
void Tune_Tcvr(double x);
int Read_Rx_Freq(void);
void Write_Rx_Freq(int freq);
/* codec.c */
gboolean Assemble_Column(unsigned char *column);
gboolean Transmit_Character(font_data_t *font_data, int glx);
gboolean Tx_FMHell_Character(font_data_t *font_data, int glx);
gboolean Tx_Feld_Character(font_data_t *font_data, int glx);
void Morse_Transmit(char *mesg);
void Make_Sin_Wavetable(void);
void Make_Cos_Wavetable(void);
  /* detect.c */
int Get_Element(void);
int Feld_Level(void);
int	FMHell_Level(void);
/* display.c */
gboolean Draw_Column(unsigned char *column);
void Display_Waterfall(void);
void Clear_Pixbuf(void);
void Clear_Wfall_Pixbuf(void);
void Print_Character(int printchr, GtkTextBuffer *text_buffer,
	 GtkWidget *scrolledwindow, GtkTextIter *iter);
void Display_Signal(int plot);
/* fft.c */
void Ifft_Init(void);
void Ifft(void);
/* main.c */
int main(int argc, char *argv[]);
gboolean Load_Config(gpointer data);
int Load_Line(char *buff, FILE *pfile);
void Usage(void);
int isFlagSet(int flag);
int isFlagClear(int flag);
void SetFlag(int flag);
void ClearFlag(int flag);
void ToggleFlag(int flag);
void Cleanup(void);
/* operation.c */
gboolean Transmit_Keybd(gpointer data);
gboolean Transmit_Macro(gpointer data);
gboolean Receive_Mode(gpointer data);
gboolean Initialize_Transmit(gpointer data);
/* sound.c */
void Setup_Sound_Card(void);
gboolean Setup_Dsp_Dev(char *dsp_dev, int mode, int *dsp_fd);
void Set_Volume_Level(int level);
void Set_Recording_Level(int level);
int Signal_Sample(void);
void DSP_Write(unsigned char *buffer, int buff_size);
/* utils.c */
gboolean Save_QSO_Record(void);
gboolean Read_QSO_Record(void);
void Clear_Record_Fields(gboolean all);
gboolean Open_Record_File(void);
void Set_TxRx_Labels(void);
gboolean Load_Fonts(char *font_file);
gboolean Find_Line(char *line, const char *label, FILE *fp);
void Find_String(char *line, int *idx);
unsigned int Hex2Bitmap(char *hex_str);
gboolean Render_Font(font_data_t *font, int glx);
gboolean Save_Pixbuf(void);
gboolean Alloc_Tx_Buff(void);
void Process_RST(GtkEditable *editable);
void Select_Macro(int macro_num);
int Open_File(FILE **fp, char *fpath, const char *mode);
void Close_File(FILE **fp);
gboolean File_Print(FILE **fp, char *string);
gboolean Locate_Line(char **line_idx, char *line, const char *string, FILE *fp);
gboolean Locate_String(char **line_idx, char *line, const char *string);
void  New_Parameters(void);

/*-------------------------------------------------------------------*/

#endif /* End of xfhell.h */

